import os
from typing import Optional

from premai import Prem

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger

logger = setup_console_logger(name="[PREMAI-GENERATOR]")


class Text2SQLGeneratorPremAI(Text2SQLGeneratorBase):
    def __init__(
        self,
        model_name: str,
        project_id: str,
        experiment_name: str,
        type: str,
        experiment_folder: Optional[str] = None,
        premai_api_key: Optional[str] = None,
        **kwargs
    ):
        self.project_id = project_id
        self.premai_api_key = premai_api_key or os.environ.get("PREMAI_API_KEY")
        self._kwargs = kwargs
        self.model_name = model_name

        super().__init__(
            experiment_name=experiment_name,
            experiment_folder=experiment_folder,
            type=type,
        )

    @property
    def load_client(self) -> Prem:
        return Prem(api_key=self.premai_api_key)

    @property
    def load_tokenizer(self) -> None:
        pass

    @property
    def model_name_or_path(self) -> str:
        return self.model_name

    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs
    ) -> str:
        prompt = data_blob["prompt"]
        max_tokens = max_new_tokens
        generation_config = {
            **kwargs,
            **{"temperature": temperature, "max_tokens": max_tokens},
        }
        generated = (
            self.client.chat.completions.create(
                project_id=self.project_id,
                messages=[{"role": "user", "content": prompt}],
                **generation_config
            )
            .choices[0]
            .message.content
        )
        return self.postprocess(output_string=generated) if postprocess else generated



from typing import Optional

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger

logger = setup_console_logger(name="[OLLAMA-GENERATOR]")

try:
    from ollama import Client
except ImportError:
    logger.warn("Ensure ollama is installed")
    logger.warn("Install Ollama: curl -fsSL https://ollama.com/install.sh | sh")
    logger.warn("Install Ollama python: pip install ollama")


class Text2SQLGeneratorOllama(Text2SQLGeneratorBase):
    def __init__(
        self, 
        model_name: str,
        experiment_name: str,
        type: str,
        experiment_folder: Optional[str]=None,
        **kwargs
    ):
        self._kwargs = kwargs
        self.model_name = model_name
        super().__init__(
            experiment_name=experiment_name,
            experiment_folder=experiment_folder,
            type=type
        )
    
    @property
    def load_client(self):
        return Client(host='http://localhost:11434')
    
    @property
    def load_tokenizer(self):
        pass

    @property
    def model_name_or_path(self):
        return self.model_name

    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs
    ) -> str:
        prompt = data_blob["prompt"]
        response = self.load_client.chat(
            model=self.model_name_or_path,
            messages=[{"role":"user", "content":prompt}],
            options=dict(
                temperature=temperature,
                num_ctx=2048 + max_new_tokens
            )
        )["message"]["content"]
        return self.postprocess(output_string=response) if postprocess else response
    


from typing import Optional

import pandas as pd
from tqdm.auto import tqdm

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.agents.base import AnalyserWorkerOutput, AnalysisWorkerBase
from premsql.agents.baseline.prompts import (
    BASELINE_ANALYSIS_MERGER_PROMPT,
    BASELINE_ANALYSIS_WORKER_PROMPT,
)
from premsql.agents.utils import convert_df_to_dict

logger = setup_console_logger("[BASELINE-ANALYSER-WORKER]")

CHUNK_TEMPLATE = """
# Analysis:
{analysis}

# Reasoning
{reasoning}
"""

# TODO: Need to think of the case when there is no df being passed


class BaseLineAnalyserWorker(AnalysisWorkerBase):
    def __init__(self, generator: Text2SQLGeneratorBase) -> None:
        self.generator = generator

    def run_chunkwise_analysis(
        self,
        question: str,
        input_dataframe: pd.DataFrame,
        chunk_size: Optional[int] = 20,
        max_chunks: Optional[int] = 20,
        temperature: Optional[float] = 0.19,
        max_new_tokens: Optional[int] = 600,
        analysis_prompt_template: Optional[str] = BASELINE_ANALYSIS_WORKER_PROMPT,
        merger_prompt_template: Optional[str] = BASELINE_ANALYSIS_MERGER_PROMPT,
        verbose: Optional[bool] = False,
    ) -> tuple[str, str]:
        num_chunks = (len(input_dataframe) + chunk_size - 1) // chunk_size
        chunks = [
            input_dataframe[i * chunk_size : (i + 1) * chunk_size]
            for i in range(num_chunks)
        ][:max_chunks]
        analysis_list = []
        num_errors = 0

        for i, chunk in tqdm(enumerate(chunks), total=len(chunks)):
            analysis, error_from_model = self.analyse(
                question=question,
                input_dataframe=chunk,
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                prompt_template=analysis_prompt_template,
            )
            if error_from_model:
                num_errors += 1
                logger.error(f"Error while analysing: {i}, Skipping ...")
                continue

            if verbose:
                logger.info(
                    CHUNK_TEMPLATE.format(
                        analysis=analysis["analysis"],
                        reasoning=analysis["analysis_reasoning"],
                    )
                )
            analysis_list.append(analysis)

        analysis_list_str = "\n".join(
            [
                analysis["analysis"] + " " + analysis["analysis_reasoning"]
                for analysis in analysis_list
            ]
        )
        if num_errors < len(chunks):
            summarized_analysis_prompt = merger_prompt_template.format(
                analysis=analysis_list_str
            )
            summary = self.generator.generate(
                data_blob={"prompt": summarized_analysis_prompt},
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                postprocess=False,
            )
            analysis = {
                "analysis": summary,
                "analysis_reasoning": "Analysis summarised by AI",
            }
            error_from_model = None

        else:
            analysis = {
                "analysis": "\n".join(
                    [
                        content["analyse"] if "analyse" in content else ""
                        for content in analysis_list
                    ]
                ),
                "analysis_reasoning": "Appending all the analysis",
            }
            error_from_model = "Model not able to summarise analysis"

        return analysis, error_from_model

    def analyse(
        self,
        question: str,
        input_dataframe: pd.DataFrame,
        temperature: Optional[float] = 0.19,
        max_new_tokens: Optional[int] = 512,
        prompt_template: Optional[str] = BASELINE_ANALYSIS_WORKER_PROMPT,
    ) -> dict:
        output = self.generator.generate(
            data_blob={
                "prompt": prompt_template.format(
                    dataframe=str(input_dataframe), question=question
                )
            },
            temperature=temperature,
            max_new_tokens=max_new_tokens,
            postprocess=False,
        )
        try:
            sections = output.split('# ')
            analysis_from_model, reasoning_from_model = '', ''
            for section in sections:
                if section.startswith('Analysis:'):
                    analysis_from_model = section.strip()
                elif section.startswith('Reasoning:'):
                    reasoning_from_model = section.strip()
            
            analysis = {
                "analysis": analysis_from_model,
                "analysis_reasoning": reasoning_from_model
            }
            error_from_model = None
        except Exception as e:
            analysis = {
                "analysis": output,
                "analysis_reasoning": "Not able to split analysis and reasoning",
            }
            error_from_model = str(e)

        logger.info(analysis)
        logger.info("------------")
        logger.info(error_from_model)

        return analysis, error_from_model

    def run(
        self,
        question: str,
        input_dataframe: pd.DataFrame,
        do_chunkwise_analysis: Optional[bool] = False,
        chunk_size: Optional[int] = 20,
        max_chunks: Optional[int] = 20,
        temperature: Optional[float] = 0.19,
        max_new_tokens: Optional[int] = 600,
        analysis_prompt_template: Optional[str] = BASELINE_ANALYSIS_WORKER_PROMPT,
        analysis_merger_template: Optional[str] = BASELINE_ANALYSIS_MERGER_PROMPT,
        verbose: Optional[bool] = False,
    ) -> AnalyserWorkerOutput:
        if len(input_dataframe) > chunk_size and do_chunkwise_analysis:
            logger.info("Going for chunk wise analysis ...")
            analysis, error_from_model = self.run_chunkwise_analysis(
                question=question,
                input_dataframe=input_dataframe,
                chunk_size=chunk_size,
                max_chunks=max_chunks,
                analysis_prompt_template=analysis_prompt_template,
                merger_prompt_template=analysis_merger_template,
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                verbose=verbose,
            )
        else:
            if len(input_dataframe) > chunk_size:
                logger.info(
                    "Truncating table, you can also choose chunk wise analysis, but it takes more time."
                )
            analysis, error_from_model = self.analyse(
                question=question,
                input_dataframe=input_dataframe.iloc[:chunk_size, :],
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                prompt_template=analysis_prompt_template,
            )
        return AnalyserWorkerOutput(
            question=question,
            input_dataframe=convert_df_to_dict(df=input_dataframe),
            analysis=analysis.get("analysis", "Not able to analyse"),
            analysis_reasoning=analysis.get("analysis_reasoning", None),
            error_from_model=error_from_model,
            additional_input={
                "temperature": temperature,
                "max_new_tokens": max_new_tokens,
                "chunkwise_analysis": do_chunkwise_analysis,
                "chunk_size": chunk_size,
                "max_chunks": max_chunks,
            },
        )



import traceback

from contextlib import asynccontextmanager
from datetime import datetime
from typing import Optional

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

from premsql.logger import setup_console_logger
from premsql.agents.base import AgentBase, AgentOutput

logger = setup_console_logger("[FASTAPI-INFERENCE-SERVICE]")


class QuestionInput(BaseModel):
    question: str


class SessionInfoResponse(BaseModel):
    status: int
    session_name: Optional[str] = None
    db_connection_uri: Optional[str] = None
    session_db_path: Optional[str] = None
    base_url: Optional[str] = None
    created_at: Optional[datetime] = None


class ChatHistoryResponse(BaseModel):
    message_id: int
    agent_output: AgentOutput


class CompletionResponse(BaseModel):
    message_id: int
    message: AgentOutput


class AgentServer:
    def __init__(
        self,
        agent: AgentBase,
        url: Optional[str] = "localhost",
        port: Optional[int] = 8100,
    ) -> None:
        self.agent = agent
        self.port = port
        self.url = url
        self.app = self.create_app()

    @asynccontextmanager
    async def lifespan(self, app: FastAPI):
        # Startup: Log the initialization
        logger.info("Starting up the application")
        yield
        # Shutdown: Clean up resources
        logger.info("Shutting down the application")
        if hasattr(self.agent, "cleanup"):
            await self.agent.cleanup()

    def create_app(self):
        app = FastAPI(lifespan=self.lifespan)
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],  # Allows all origins
            allow_credentials=True,
            allow_methods=["*"],  # Allows all methods
            allow_headers=["*"],  # Allows all headers
        )

        @app.post("/completion", response_model=CompletionResponse)
        async def completion(input_data: QuestionInput):
            try:
                result = self.agent(question=input_data.question, server_mode=True)
                message_id = self.agent.history.get_latest_message_id()
                return CompletionResponse(
                    message=AgentOutput(**result.model_dump()), message_id=message_id
                )
            except Exception as e:
                stack_trace = traceback.format_exc()
                logger.error(stack_trace)
                logger.error(f"Error processing query: {str(e)}")
                raise HTTPException(
                    status_code=500, detail=f"Error processing query: {str(e)}"
                )

        # TODO: I need a method which will just get the "latets message_id"

        @app.get("/chat_history/{message_id}", response_model=ChatHistoryResponse)
        async def get_chat_history(message_id: int):
            try:
                exit_output = self.agent.history.get_by_message_id(
                    message_id=message_id
                )
                if exit_output is None:
                    raise HTTPException(
                        status_code=404,
                        detail=f"Message with ID {message_id} not found",
                    )
                agent_output = self.agent.convert_exit_output_to_agent_output(
                    exit_output=exit_output
                )
                return ChatHistoryResponse(
                    message_id=message_id, agent_output=agent_output
                )
            except Exception as e:
                logger.error(f"Error retrieving chat history: {str(e)}")
                raise HTTPException(
                    status_code=500, detail=f"Error retrieving chat history: {str(e)}"
                )

        @app.get("/")
        async def health_check():
            return {
                "status_code": 200, 
                "status": f"healthy, running: {self.agent.session_name}"
            }

        @app.get("/health")
        async def health_check():
            return {"status_code": 200, "status": "healthy"}

        @app.get("/session_info", response_model=SessionInfoResponse)
        async def get_session_info():
            try:
                session_name = getattr(self.agent, "session_name", None)
                db_connection_uri = getattr(self.agent, "db_connection_uri", None)
                session_db_path = getattr(self.agent, "session_db_path", None)

                if any(
                    attr is None
                    for attr in [session_name, db_connection_uri, session_db_path]
                ):
                    raise ValueError("One or more required attributes are None")

                return SessionInfoResponse(
                    status=200,
                    session_name=session_name,
                    db_connection_uri=db_connection_uri,
                    session_db_path=session_db_path,
                    base_url=f"{self.url}:{self.port}",
                    created_at=datetime.now(),
                )
            except Exception as e:
                logger.error(f"Error getting session info: {str(e)}")
                return SessionInfoResponse(
                    status=500,
                    session_name=None,
                    db_connection_uri=None,
                    session_db_path=None,
                    base_url=None,
                    created_at=None,
                )

        return app

    def launch(self):
        import uvicorn

        logger.info(f"Starting server on port {self.port}")
        uvicorn.run(self.app, host=self.url, port=int(self.port))



import streamlit as st
from premsql.playground.backend.backend_client import BackendAPIClient
from premsql.playground.backend.api.pydantic_models import SessionCreationRequest

additional_link_markdown = """
Here are some quick links to get you started with Prem:

- Head over to [Prem App](https://app.premai.io/projects/) to start building on Gen AI.
- [Prem AI documentation](https://docs.premai.io/get-started/why-prem)
- [PremSQL documentation](https://docs.premai.io/premsql/introduction)
"""

class SessionComponent:
    def __init__(self) -> None:
        self.backend_client = BackendAPIClient()
    
    def render_list_sessions(self):
        with st.sidebar:
            st.sidebar.title("Your Past Sessions")
            all_sessions = self.backend_client.list_sessions(page_size=100).sessions
            if all_sessions:
                all_sessions = [session.session_name for session in all_sessions]

                selected_session = st.selectbox(
                    label="Your Sessions (refresh if you have created a new one)",
                    options=all_sessions,
                )
                return selected_session
    
    def render_register_session(self):
        with st.sidebar:
            st.sidebar.title("Register new Session")
            with st.form(
                key="session_creation",
                clear_on_submit=True,
                border=100,
                enter_to_submit=False,
            ):
                base_url = st.text_input(
                    label="base_url",
                    placeholder="the base url in which AgentServer is running"
                )
                button = st.form_submit_button(label="Submit")
            if button:
                response = self.backend_client.create_session(
                    request=SessionCreationRequest(base_url=base_url)
                )
                if response.status_code == 500:
                    st.toast(body=st.markdown(response.error_message), icon="❌")
                else:
                    st.toast(body=f"Session: {response.session_name} created successfully", icon="🥳")
                return response
    
    def render_additional_links(self):
        with st.sidebar:
            with st.container(height=200):
                st.markdown(additional_link_markdown)

    
    def render_delete_session_view(self):
        with st.sidebar:
            with st.expander(label="Delete a session"):
                with st.form(key="delete_session", clear_on_submit=True, enter_to_submit=False):
                    session_name = st.text_input(label="Enter session name")
                    button = st.form_submit_button(label="Submit")
                    if button:
                        all_sessions = self.backend_client.list_sessions(page_size=100).sessions
                        all_sessions = [session.session_name for session in all_sessions]
                        if session_name not in all_sessions:
                            st.error("Session does not exist")
                        else:
                            self.backend_client.delete_session(session_name=session_name)
                            st.success(f"Deleted session: {session_name}. Please refresh")


from abc import ABC, abstractmethod

import numpy as np


class BaseExecutor(ABC):
    @abstractmethod
    def execute_sql(self, sql: str, dsn_or_db_path: str) -> dict:
        return {"result": None, "execution_time": None, "error": None}

    def match_sqls(
        self, predicted_sql: str, gold_sql: str, dsn_or_db_path: str
    ) -> bool:
        prediction = self.execute_sql(sql=predicted_sql, dsn_or_db_path=dsn_or_db_path)
        gold = self.execute_sql(sql=gold_sql, dsn_or_db_path=dsn_or_db_path)
        if prediction["error"]:
            return {
                "result": 0,
                "error": prediction["error"],
            }

        is_match = set(prediction["result"]) == set(gold["result"])
        return {
            "result": int(is_match),
            "error": None if is_match else "Table mismatch",
        }

    def clean_abnormal(self, input: list[float]) -> list[float]:
        input_array = np.asarray(input)
        mean = np.mean(input_array)
        std = np.std(input_array)
        return [x for x in input_array if mean - 3 * std < x < mean + 3 * std]

    def iterated_execution(
        self,
        predicted_sql: str,
        gold_sql: str,
        dsn_or_db_path: str,
        num_iterations: int,
    ) -> dict:
        is_match = self.match_sqls(
            predicted_sql=predicted_sql,
            gold_sql=gold_sql,
            dsn_or_db_path=dsn_or_db_path,
        )

        if is_match["result"] == 1:
            diff_list = [
                self.execute_sql(sql=gold_sql, dsn_or_db_path=dsn_or_db_path)[
                    "execution_time"
                ]
                / self.execute_sql(sql=gold_sql, dsn_or_db_path=dsn_or_db_path)[
                    "execution_time"
                ]
                for _ in range(num_iterations)
            ]
            processed_diff_list = self.clean_abnormal(diff_list)
            return {
                "result": sum(processed_diff_list) / len(processed_diff_list),
                "error": None,
            }
        return {"result": 0, "error": is_match["error"]}



from textwrap import dedent
from typing import Literal, Optional

from premsql.executors.base import BaseExecutor
from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.agents.base import Text2SQLWorkerBase
from premsql.agents.baseline.prompts import (
    BASELINE_TEXT2SQL_TABLE_SELECTION_PROMPT,
    BASELINE_TEXT2SQL_WORKER_ERROR_HANDLING_PROMPT,
    BASELINE_TEXT2SQL_WORKER_PROMPT,
    BASELINE_TEXT2SQL_WORKER_PROMPT_NO_FEWSHOT,
)
from premsql.agents.models import Text2SQLWorkerOutput
from premsql.agents.utils import execute_and_render_result

logger = setup_console_logger("[BASELINE-TEXT2SQL-WORKER]")


class BaseLineText2SQLWorker(Text2SQLWorkerBase):
    def __init__(
        self,
        db_connection_uri: str,
        generator: Text2SQLGeneratorBase,
        helper_model: Optional[Text2SQLGeneratorBase] = None,
        executor: Optional[BaseExecutor] = None,
        include_tables: Optional[list] = None,
        exclude_tables: Optional[list] = None,
        auto_filter_tables: Optional[bool] = False,
    ):
        super().__init__(
            db_connection_uri=db_connection_uri,
            generator=generator,
            executor=executor,
            include_tables=include_tables,
            exclude_tables=exclude_tables,
        )

        self.corrector = helper_model
        self.table_filer_worker = helper_model
        self.auto_filter_tables = auto_filter_tables

    @staticmethod
    def show_dataframe(output: Text2SQLWorkerOutput):
        import pandas as pd

        if output.output_dataframe:
            df = pd.DataFrame(
                output.output_dataframe["data"],
                columns=output.output_dataframe["columns"],
            )
            return df
        return pd.DataFrame({})

    def filer_tables_from_schema(
        self, question: str, additional_input: Optional[str] = None
    ) -> dict:
        prompt = BASELINE_TEXT2SQL_TABLE_SELECTION_PROMPT.format(
            schema=self.db.get_context()["table_info"],
            additional_info=additional_input,
            question=question,
        )
        all_tables = self.db.get_usable_table_names()
        try:
            to_include = []
            output = self.corrector.generate({"prompt": prompt}, postprocess=False)
            output = eval(output)
            for table in all_tables:
                if table in output["include"]:
                    to_include.append(table)
        except Exception as e:
            logger.info(f"Error while selecting table: {e}")
            to_include = all_tables
        return to_include

    def _create_prompt(
        self,
        question: str,
        additional_knowledge: Optional[str] = None,
        fewshot_dict: Optional[dict] = None,
        prompt_template: Optional[str] = BASELINE_TEXT2SQL_WORKER_PROMPT,
    ) -> str:
        # Initialize database first
        self.db = self.initialize_database(
            db_connection_uri=self.db_connection_uri, include_tables=None
        )
        
        # Get tables to include
        to_include = []
        if self.auto_filter_tables:
            try:
                to_include = self.filer_tables_from_schema(
                    question=question, additional_input=additional_knowledge
                )
                logger.info(f"Selected tables in schema: {to_include}")
            except Exception as e:
                logger.warning(f"Error filtering tables: {e}. Using all tables.")
        
        # Get schema information
        schema_prompt = self.db.get_context()["table_info"]
        
        # Build knowledge prompt
        knowledge_prompt = ""
        if fewshot_dict:
            template = dedent("""
                Question: {question}
                SQL: {sql}
                """)
            try:
                knowledge_prompt = "".join(
                    template.format(question=sample_question, sql=sample_sql)
                    for sample_question, sample_sql in fewshot_dict.items()
                )
            except Exception as e:
                logger.warning(f"Error formatting fewshot examples: {e}")
        elif to_include and len(to_include) <= 10:
            try:
                self.db._sample_rows_in_table_info = 3
                knowledge_prompt = str(self.db.get_table_info(table_names=to_include))
            except Exception as e:
                logger.warning(f"Error getting table info: {e}")

        try:
            if fewshot_dict or (to_include and len(to_include) <= 10):
                prompt = prompt_template.format(
                    schemas=schema_prompt if fewshot_dict else "",
                    additional_knowledge=additional_knowledge or "",
                    few_shot_examples=knowledge_prompt,
                    question=question,
                )
            else:
                prompt = BASELINE_TEXT2SQL_WORKER_PROMPT_NO_FEWSHOT.format(
                    schemas=schema_prompt,
                    additional_knowledge=additional_knowledge or "",
                    question=question,
                )
            return prompt
        except Exception as e:
            logger.error(f"Error formatting prompt: {e}")
            raise

    def run(
        self,
        question: str,
        additional_knowledge: Optional[str] = None,
        fewshot_dict: Optional[dict] = None,
        temperature: Optional[float] = 0.1,
        max_new_tokens: Optional[int] = 256,
        render_results_using: Optional[Literal["json", "dataframe"]] = "json",
        prompt_template: Optional[str] = BASELINE_TEXT2SQL_WORKER_PROMPT,
        error_handling_prompt_template: Optional[
            str
        ] = BASELINE_TEXT2SQL_WORKER_ERROR_HANDLING_PROMPT,
        **kwargs,
    ) -> Text2SQLWorkerOutput:
        if question.startswith("`") and question.endswith("`"):
            result = execute_and_render_result(
                db=self.db, sql=question.replace('`', ''), using=render_results_using
            ) 
            return Text2SQLWorkerOutput(
            db_connection_uri=self.db_connection_uri,
            sql_string=question.startswith("`"),
            sql_reasoning=None,
            input_dataframe=None,
            output_dataframe=result["dataframe"],  # Truncating to
            question=question,
            error_from_model=result["error_from_model"],
            additional_input={
                "additional_knowledge": additional_knowledge,
                "fewshot_dict": fewshot_dict,
                "temperature": temperature,
                "max_new_tokens": max_new_tokens,
                **kwargs,
            },
        )


        prompt = self._create_prompt(
            question=question,
            additional_knowledge=additional_knowledge,
            fewshot_dict=fewshot_dict,
            prompt_template=prompt_template,
        )
        generated_sql = self.generator.execution_guided_decoding(
            data_blob={"prompt": prompt, "db_path": self.db_connection_uri},
            executor=self.executor,
            temperature=temperature,
            max_new_tokens=max_new_tokens,
            max_retries=5,
            postprocess=True,
            **kwargs,
        )

        result = execute_and_render_result(
            db=self.db, sql=generated_sql, using=render_results_using
        )

        if result["error_from_model"] is not None:
            logger.info("=> Going for final correction ...")
            generated_sql = self.do_correction(
                question=question,
                result=result,
                additional_knowledge=additional_knowledge,
                fewshot_dict=fewshot_dict,
                prompt_template=prompt_template,
                error_handling_prompt_template=error_handling_prompt_template,
                **kwargs,
            )
            result = execute_and_render_result(
                db=self.db, sql=generated_sql, using=render_results_using
            )

        return Text2SQLWorkerOutput(
            db_connection_uri=self.db_connection_uri,
            sql_string=generated_sql,
            sql_reasoning=None,
            input_dataframe=None,
            output_dataframe=result["dataframe"],  # Truncating to
            question=question,
            error_from_model=result["error_from_model"],
            additional_input={
                "additional_knowledge": additional_knowledge,
                "fewshot_dict": fewshot_dict,
                "temperature": temperature,
                "max_new_tokens": max_new_tokens,
                **kwargs,
            },
        )

    def do_correction(
        self,
        question: str,
        result: dict,
        additional_knowledge: Optional[str] = None,
        fewshot_dict: Optional[dict] = None,
        prompt_template: Optional[str] = BASELINE_TEXT2SQL_WORKER_PROMPT,
        error_handling_prompt_template: Optional[
            str
        ] = BASELINE_TEXT2SQL_WORKER_ERROR_HANDLING_PROMPT,
        **kwargs,
    ):
        if not self.corrector:
            logger.info("Corrector model not defined, no furthur correction possible.")

        error_prompt = error_handling_prompt_template.format(
            existing_prompt=self._create_prompt(
                question=question,
                additional_knowledge=additional_knowledge,
                fewshot_dict=fewshot_dict,
                prompt_template=prompt_template,
            ),
            error_msg=result["error_from_model"],
            sql=result["sql_string"],
        )
        return self.generator.generate(
            data_blob={"prompt": error_prompt}, postprocess=True, **kwargs
        )



import io
from typing import Callable, Dict

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from PIL import Image

from premsql.logger import setup_console_logger
from premsql.agents.tools.plot.base import BasePlotTool

logger = setup_console_logger("[MATPLOTLIB-TOOL]")


class SimpleMatplotlibTool(BasePlotTool):
    def __init__(self):
        self.plot_functions: Dict[
            str, Callable[[pd.DataFrame, str, str, Axes], None]
        ] = {
            "area": self._area_plot,
            "bar": self._bar_plot,
            "scatter": self._scatter_plot,
            "histogram": self._histogram_plot,
            "line": self._line_plot,
        }

    def run(self, data: pd.DataFrame, plot_config: Dict[str, str]) -> Figure:
        try:
            self._validate_config(data, plot_config)

            plot_type = plot_config["plot_type"]
            x = plot_config["x"]
            y = plot_config["y"]

            fig, ax = plt.subplots(figsize=(10, 6))
            self.plot_functions[plot_type](data, x, y, ax)

            plt.title(f"{plot_type.capitalize()} Plot: {x} vs {y}")
            plt.xlabel(x)
            plt.ylabel(y)
            plt.tight_layout()

            return fig
        except Exception as e:
            logger.error(f"Error creating plot: {str(e)}")
            return plt.figure()  # Return an empty figure on error

    def _validate_config(self, df: pd.DataFrame, plot_config: Dict[str, str]) -> None:
        required_keys = ["plot_type", "x", "y"]
        missing_keys = [key for key in required_keys if key not in plot_config]
        if missing_keys:
            raise ValueError(
                f"Missing required keys in plot_config: {', '.join(missing_keys)}"
            )

        if plot_config["x"] not in df.columns:
            raise ValueError(f"Column '{plot_config['x']}' not found in DataFrame")

        if plot_config["y"] not in df.columns:
            raise ValueError(f"Column '{plot_config['y']}' not found in DataFrame")

        if plot_config["plot_type"] not in self.plot_functions:
            raise ValueError(f"Unsupported plot type: {plot_config['plot_type']}")

    def _area_plot(self, df: pd.DataFrame, x: str, y: str, ax: Axes) -> None:
        ax.fill_between(df[x], df[y])

    def _bar_plot(self, df: pd.DataFrame, x: str, y: str, ax: Axes) -> None:
        ax.bar(df[x], df[y])

    def _scatter_plot(self, df: pd.DataFrame, x: str, y: str, ax: Axes) -> None:
        ax.scatter(df[x], df[y])

    def _histogram_plot(self, df: pd.DataFrame, x: str, y: str, ax: Axes) -> None:
        ax.hist(df[x], bins=20)

    def _line_plot(self, df: pd.DataFrame, x: str, y: str, ax: Axes) -> None:
        ax.plot(df[x], df[y])

    def convert_plot_to_image(self, fig: Figure) -> Image.Image:
        buf = io.BytesIO()
        fig.savefig(buf, format="png")
        buf.seek(0)
        return Image.open(buf)



import os
import tempfile
import sqlite3
from platformdirs import user_cache_dir
from typing import List, Literal, Optional

from premsql.logger import setup_console_logger
from premsql.agents.models import ExitWorkerOutput
from premsql.agents.utils import convert_exit_output_to_agent_output

logger = setup_console_logger("[PIPELINE-MEMORY]")



class AgentInteractionMemory:
    def __init__(self, session_name: str, db_path: Optional[str] = None):
        self.session_name = session_name
        self.db_path = db_path or os.path.join(
            os.getcwd(), "premsql", "premsql_pipeline_memory.db"
        )

        logger.info(self.db_path)
        os.makedirs(os.path.dirname(self.db_path), exist_ok=True)
        self.conn = sqlite3.connect(self.db_path)
        self.create_table_if_not_exists()


    def list_sessions(self) -> List[str]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()
        return [table[0] for table in tables if table[0] != "sqlite_sequence"]

    def create_table_if_not_exists(self):
        cursor = self.conn.cursor()
        cursor.execute(
            f"""
        CREATE TABLE IF NOT EXISTS {self.session_name} (
            message_id INTEGER PRIMARY KEY AUTOINCREMENT,
            question TEXT,
            db_connection_uri TEXT,
            route_taken TEXT,
            sql_string TEXT,
            sql_reasoning TEXT,
            sql_input_dataframe TEXT,
            sql_output_dataframe TEXT,
            error_from_sql_worker TEXT,
            analysis TEXT,
            analysis_reasoning TEXT,
            analysis_input_dataframe TEXT,
            error_from_analysis_worker TEXT,
            plot_config TEXT,
            plot_input_dataframe TEXT,
            plot_output_dataframe TEXT,
            image_to_plot TEXT,
            plot_reasoning TEXT,
            error_from_plot_worker TEXT,
            followup_route_to_take TEXT,
            followup_suggestion TEXT,
            error_from_followup_worker TEXT,
            additional_input TEXT,
            timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
        )
        """
        )
        self.conn.commit()

    def get(
        self,
        limit: Optional[int] = None,
        order: Optional[Literal["DESC", "ASC"]] = "DESC",
    ) -> List[tuple[int, ExitWorkerOutput]]:
        cursor = self.conn.cursor()
        query = f"SELECT * FROM {self.session_name} ORDER BY message_id {order}"
        if limit is not None:
            query += " LIMIT ?"
            cursor.execute(query, (limit,))
        else:
            cursor.execute(query)
        rows = cursor.fetchall()
        return [
            {"message_id": row[0], "message": self._row_to_exit_worker_output(row)}
            for row in rows
        ]

    def get_latest_message_id(self) -> Optional[int]:
        cursor = self.conn.cursor()
        query = f"SELECT message_id FROM {self.session_name} ORDER BY message_id DESC LIMIT 1"
        cursor.execute(query)
        row = cursor.fetchone()
        return row[0] if row else None
    
    def generate_messages_from_session(
            self, session_name: str, limit: int = 100, server_mode: bool=False
        ):
        cursor = self.conn.cursor()
        query = f"SELECT * FROM {session_name} ORDER BY message_id ASC LIMIT {limit}"
        cursor.execute(query)
        rows = cursor.fetchall()
        for row in rows:
            yield self._row_to_exit_worker_output(row=row) if server_mode == False else convert_exit_output_to_agent_output(
                self._row_to_exit_worker_output(row=row)
            )

    def get_by_message_id(self, message_id: int) -> Optional[dict]:
        cursor = self.conn.cursor()
        query = f"SELECT * FROM {self.session_name} WHERE message_id = ?"
        cursor.execute(query, (message_id,))
        row = cursor.fetchone()
        if row is None:
            return None
        return self._row_to_exit_worker_output(row=row)

    def push(self, output: ExitWorkerOutput):
        cursor = self.conn.cursor()
        cursor.execute(
            f"""
        INSERT INTO {self.session_name} (
            question, db_connection_uri, route_taken, sql_string, sql_reasoning,
            sql_input_dataframe, sql_output_dataframe, error_from_sql_worker,
            analysis, analysis_reasoning, analysis_input_dataframe,
            error_from_analysis_worker, plot_config, plot_input_dataframe,
            plot_output_dataframe, image_to_plot, plot_reasoning,
            error_from_plot_worker, followup_route_to_take, followup_suggestion,
            error_from_followup_worker, additional_input
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
        """,
            self._exit_worker_output_to_tuple(output),
        )
        self.conn.commit()
        logger.info("Pushed to the database")

    def delete_table(self):
        cursor = self.conn.cursor()
        try:
            cursor.execute(f"DROP TABLE IF EXISTS {self.session_name}")
            self.conn.commit()
            logger.info(f"Table '{self.session_name}' has been deleted.")
        except sqlite3.Error as e:
            logger.error(f"Error deleting table '{self.session_name}': {e}")
        finally:
            cursor.close()

    def _row_to_exit_worker_output(self, row) -> ExitWorkerOutput:
        try:
            return ExitWorkerOutput(
                session_name=self.session_name,
                question=row[1],
                db_connection_uri=row[2],
                route_taken=row[3],
                sql_string=row[4],
                sql_reasoning=row[5],
                sql_input_dataframe=self._parse_json(row[6]),
                sql_output_dataframe=self._parse_json(row[7]),
                error_from_sql_worker=row[8],
                analysis=row[9],
                analysis_reasoning=row[10],
                analysis_input_dataframe=self._parse_json(row[11]),
                error_from_analysis_worker=row[12],
                plot_config=self._parse_json(row[13]),
                plot_input_dataframe=self._parse_json(row[14]),
                plot_output_dataframe=self._parse_json(row[15]),
                image_to_plot=row[16],
                plot_reasoning=row[17],
                error_from_plot_worker=row[18],
                followup_route_to_take=row[19],
                followup_suggestion=row[20],
                error_from_followup_worker=row[21],
                additional_input=self._parse_json(row[22]),
            )
        except Exception as e:
            logger.error(f"Error converting row to ExitWorkerOutput: {e}")
            return None

    def _exit_worker_output_to_tuple(self, output: ExitWorkerOutput) -> tuple:
        return (
            output.question,
            output.db_connection_uri,
            output.route_taken,
            output.sql_string,
            output.sql_reasoning,
            self._serialize_json(output.sql_input_dataframe),
            self._serialize_json(output.sql_output_dataframe),
            output.error_from_sql_worker,
            output.analysis,
            output.analysis_reasoning,
            self._serialize_json(output.analysis_input_dataframe),
            output.error_from_analysis_worker,
            self._serialize_json(output.plot_config),
            self._serialize_json(output.plot_input_dataframe),
            self._serialize_json(output.plot_output_dataframe),
            output.image_to_plot,
            output.plot_reasoning,
            output.error_from_plot_worker,
            output.followup_route_to_take,
            output.followup_suggestion,
            output.error_from_followup_worker,
            self._serialize_json(output.additional_input),
        )

    def _parse_json(self, json_str):
        import json

        if not json_str:
            return None
        try:
            return json.loads(json_str)
        except json.JSONDecodeError:
            logger.warning(f"Failed to parse JSON: {json_str}")
            return None

    def _serialize_json(self, obj):
        import json

        if obj is None:
            return None
        try:
            return json.dumps(obj)
        except TypeError:
            logger.warning(f"Failed to serialize object to JSON: {obj}")
            return None

    def clear(self):
        cursor = self.conn.cursor()
        cursor.execute(f"DELETE FROM {self.session_name}")
        self.conn.commit()

    def close(self):
        self.conn.close()

    def __del__(self):
        self.close()

    def delete_table(self):
        cursor = self.conn.cursor()
        try:
            cursor.execute(f"DROP TABLE IF EXISTS {self.session_name}")
            self.conn.commit()
            logger.info(f"Table '{self.session_name}' has been deleted.")
        except sqlite3.Error as e:
            logger.error(f"Error deleting table '{self.session_name}': {e}")
        finally:
            cursor.close()

    def get_latest_dataframe(
        self, decision: Literal["plot", "analyse", "query", "followup"]
    ) -> dict:
        contents = self.get(limit=1)
        if not contents:
            return {}

        _, content = contents[0]
        if decision == "plot" and content.plot_input_dataframe:
            return content.plot_input_dataframe
        elif decision == "analyse" and content.analysis_input_dataframe:
            return content.analysis_input_dataframe
        elif decision in ("query", "followup") and content.sql_output_dataframe:
            return content.sql_output_dataframe

        return {}



from rest_framework import serializers


class AgentOutputSerializer(serializers.Serializer):
    session_name = serializers.CharField()
    question = serializers.CharField()
    db_connection_uri = serializers.CharField()
    route_taken = serializers.ChoiceField(
        choices=["plot", "analyse", "query", "followup"]
    )
    input_dataframe = serializers.DictField(allow_null=True)
    output_dataframe = serializers.DictField(allow_null=True)
    sql_string = serializers.CharField(allow_null=True)
    analysis = serializers.CharField(allow_null=True)
    reasoning = serializers.CharField(allow_null=True)
    plot_config = serializers.DictField(allow_null=True)
    image_to_plot = serializers.CharField(allow_null=True)
    followup_route = serializers.ChoiceField(
        choices=["plot", "analyse", "query", "followup"], allow_null=True
    )
    followup_suggestion = serializers.CharField(allow_null=True)
    error_from_pipeline = serializers.CharField(allow_null=True)


# Sessions
class SessionCreationRequestSerializer(serializers.Serializer):
    base_url = serializers.CharField()


class SessionCreationResponseSerializer(serializers.Serializer):
    status_code = serializers.ChoiceField(choices=[200, 500])
    status = serializers.ChoiceField(choices=["success", "error"])

    session_id = serializers.IntegerField(allow_null=True)
    session_name = serializers.CharField(allow_null=True)
    db_connection_uri = serializers.CharField(allow_null=True)
    session_db_path = serializers.CharField(allow_null=True)
    created_at = serializers.DateTimeField(allow_null=True)
    error_message = serializers.CharField(allow_null=True)


class SessionSummarySerializer(serializers.Serializer):
    session_id = serializers.IntegerField()
    session_name = serializers.CharField(max_length=255)
    created_at = serializers.DateTimeField()
    base_url = serializers.CharField()
    db_connection_uri = serializers.CharField()
    session_db_path = serializers.CharField()


class SessionListResponseSerializer(serializers.Serializer):
    status_code = serializers.ChoiceField(choices=[200, 500])
    status = serializers.ChoiceField(choices=["success", "error"])
    sessions = SessionSummarySerializer(many=True, allow_null=True)
    total_count = serializers.IntegerField(allow_null=True)
    page = serializers.IntegerField(allow_null=True)
    page_size = serializers.IntegerField(allow_null=True)
    error_message = serializers.CharField(allow_null=True)


class SessionDeletionResponse(serializers.Serializer):
    session_name = serializers.CharField(max_length=255)
    status_code = serializers.ChoiceField(choices=[200, 404, 500])
    status = serializers.ChoiceField(choices=["success", "error"])
    error_message = serializers.CharField(allow_null=True)


# Chats (Completions)
class CompletionCreationRequestSerializer(serializers.Serializer):
    session_name = serializers.CharField()
    question = serializers.CharField()


class CompletionCreationResponseSerializer(serializers.Serializer):
    status_code = serializers.ChoiceField(choices=[200, 500])
    status = serializers.ChoiceField(choices=["success", "error"])
    message_id = serializers.IntegerField(allow_null=True)
    session_name = serializers.CharField(allow_null=True)
    message = message = AgentOutputSerializer(allow_null=True)
    created_at = serializers.DateTimeField(allow_null=True)
    question = serializers.CharField(allow_null=True)
    error_message = serializers.CharField(allow_null=True)


class CompletionSummarySerializer(serializers.Serializer):
    message_id = serializers.IntegerField()
    session_name = serializers.CharField()
    base_url = serializers.CharField()
    created_at = serializers.DateTimeField()
    question = serializers.CharField(allow_null=True)


class CompletionListResponseSerializer(serializers.Serializer):
    status_code = serializers.ChoiceField(choices=[200, 500])
    status = serializers.ChoiceField(choices=["success", "error"])
    completions = CompletionSummarySerializer(many=True, allow_null=True)
    total_count = serializers.IntegerField(allow_null=True)
    error_message = serializers.CharField(allow_null=True)


# Utility function for creating model serializers
def create_model_serializer(model_class):
    class ModelSerializer(serializers.ModelSerializer):
        class Meta:
            model = model_class
            fields = "__all__"

    return ModelSerializer



import os
import subprocess
import sys
from pathlib import Path

import click

@click.group()
@click.version_option()
def cli():
    """PremSQL CLI to manage API servers and Streamlit app"""
    pass

@cli.group()
def launch():
    """Launch PremSQL services"""
    pass


@launch.command(name='all')
def launch_all():
    """Launch both API server and Streamlit app"""
    premsql_path = Path(__file__).parent.parent.absolute()
    env = os.environ.copy()
    env["PYTHONPATH"] = str(premsql_path)
    
    # Start API server
    manage_py_path = premsql_path / "premsql" / "playground" / "backend" / "manage.py"
    if not manage_py_path.exists():
        click.echo(f"Error: manage.py not found at {manage_py_path}", err=True)
        sys.exit(1)

    # Run migrations first
    click.echo("Running database migrations...")
    try:
        subprocess.run([sys.executable, str(manage_py_path), "makemigrations"], env=env, check=True)
        subprocess.run([sys.executable, str(manage_py_path), "migrate"], env=env, check=True)
    except subprocess.CalledProcessError as e:
        click.echo(f"Error running migrations: {e}", err=True)
        sys.exit(1)

    click.echo("Starting the PremSQL backend API server...")
    subprocess.Popen([sys.executable, str(manage_py_path), "runserver"], env=env)

    # Launch the streamlit app
    click.echo("Starting the PremSQL Streamlit app...")
    main_py_path = premsql_path / "premsql" / "playground" / "frontend" / "main.py"
    if not main_py_path.exists():
        click.echo(f"Error: main.py not found at {main_py_path}", err=True)
        sys.exit(1)

    cmd = [sys.executable, "-m", "streamlit", "run", str(main_py_path), "--server.maxUploadSize=500"]
    try:
        subprocess.run(cmd, env=env, check=True)
    except KeyboardInterrupt:
        click.echo("Stopping all services...")
        stop()

@launch.command(name='api')
def launch_api():
    """Launch only the API server"""
    premsql_path = Path(__file__).parent.parent.absolute()
    env = os.environ.copy()
    env["PYTHONPATH"] = str(premsql_path)
    manage_py_path = premsql_path / "premsql" / "playground" / "backend" / "manage.py"

    if not manage_py_path.exists():
        click.echo(f"Error: manage.py not found at {manage_py_path}", err=True)
        sys.exit(1)

    # Run makemigrations
    click.echo("Running database migrations...")
    try:
        subprocess.run([sys.executable, str(manage_py_path), "makemigrations"], env=env, check=True)
        subprocess.run([sys.executable, str(manage_py_path), "migrate"], env=env, check=True)
    except subprocess.CalledProcessError as e:
        click.echo(f"Error running migrations: {e}", err=True)
        sys.exit(1)

    click.echo("Starting the PremSQL backend API server...")
    cmd = [sys.executable, str(manage_py_path), "runserver"]
    try:
        subprocess.run(cmd, env=env, check=True)
    except KeyboardInterrupt:
        click.echo("API server stopped.")

@cli.command()
def stop():
    """Stop all PremSQL services"""
    click.echo("Stopping all PremSQL services...")

    try:
        if sys.platform == "win32":
            subprocess.run(
                ["taskkill", "/F", "/IM", "python.exe", "/FI", "WINDOWTITLE eq premsql*"],
                check=True,
            )
        else:
            subprocess.run(["pkill", "-f", "manage.py runserver"], check=True)
            subprocess.run(["pkill", "-f", "streamlit"], check=True)
        click.echo("All services stopped successfully.")
    except subprocess.CalledProcessError:
        click.echo("No running services found.")
    except Exception as e:
        click.echo(f"Error stopping services: {e}", err=True)
        sys.exit(1)

if __name__ == "__main__":
    cli()


import os
from typing import Optional

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger

logger = setup_console_logger(name="[MLX-GENERATOR]")

try:
    from mlx_lm import generate
    from mlx_lm.tokenizer_utils import load_tokenizer
    from mlx_lm.utils import get_model_path, load_model
except ImportError as e:
    logger.error("Install mlx using: pip install mlx mlx-lm")



class Text2SQLGeneratorMLX(Text2SQLGeneratorBase):
    def __init__(
        self,
        model_name_or_path: str,
        experiment_name: str,
        type: str,
        experiment_folder: Optional[str] = None,
        hf_token: Optional[str] = None,
        **kwargs
    ):
        self.hf_api_key = os.environ.get("HF_TOKEN") or hf_token
        self._kwargs = kwargs
        self.mlx_model_name_or_path = model_name_or_path
        super().__init__(
            experiment_name=experiment_name,
            experiment_folder=experiment_folder,
            type=type,
        )

    @property
    def load_client(self):
        model_path = get_model_path(self.model_name_or_path)
        model = load_model(model_path, **self._kwargs)
        return model

    @property
    def load_tokenizer(self):
        model_path = get_model_path(self.model_name_or_path)
        return load_tokenizer(model_path, **self._kwargs)

    @property
    def model_name_or_path(self):
        return self.mlx_model_name_or_path

    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs
    ) -> str:
        prompt = data_blob["prompt"]
        temp = temperature
        generation_args = {"temp": temp, **kwargs}
        output = generate(
            model=self.client,
            tokenizer=self.tokenizer,
            prompt=prompt,
            max_tokens=max_new_tokens,
            **generation_args
        )
        return self.postprocess(output) if postprocess else output



from pathlib import Path
from typing import Optional, Union

from datasets import load_dataset
from tqdm.auto import tqdm

from premsql.datasets.base import (
    SupervisedDatasetForTraining,
    Text2SQLBaseDataset,
    Text2SQLBaseInstance,
)
from premsql.logger import setup_console_logger
from premsql.prompts import BASE_TEXT2SQL_PROMPT
from premsql.utils import filter_options, save_to_json

logger = setup_console_logger("[GRETELAI-DATASET]")


class GretelAIInstance(Text2SQLBaseInstance):
    def __init__(self, dataset: list[dict]) -> None:
        super().__init__(dataset)

    def apply_prompt(
        self,
        num_fewshot: Optional[int] = None,
        prompt_template: Optional[str] = BASE_TEXT2SQL_PROMPT,
    ):
        prompt_template = (
            BASE_TEXT2SQL_PROMPT if prompt_template is None else prompt_template
        )
        for blob in tqdm(self.dataset, total=len(self.dataset), desc="Applying prompt"):
            few_shot_prompt = (
                ""
                if num_fewshot is None
                else self.add_few_shot_examples(db_id=blob["db_id"], k=num_fewshot)
            )
            final_prompt = prompt_template.format(
                schemas=blob["context"],
                additional_knowledge="",
                few_shot_examples=few_shot_prompt,
                question=blob["question"],
            )
            blob["prompt"] = final_prompt
        return self.dataset


class GretelAIDataset(Text2SQLBaseDataset):
    def __init__(
        self,
        split: Optional[str] = "train",
        dataset_folder: Optional[Union[str, Path]] = "./data",
        hf_token: Optional[str] = None,
        force_download: Optional[bool] = False,
    ):
        dataset_folder = Path(dataset_folder)
        dataset_path = dataset_folder / "gretel"
        if not dataset_path.exists() or force_download:
            dataset_path.mkdir(parents=True, exist_ok=True)
            dataset = []
            raw_dataset = load_dataset("gretelai/synthetic_text_to_sql", token=hf_token)

            for split in ["train", "test"]:
                for content in raw_dataset[split]:
                    blob_content = {
                        "id": content["id"],
                        "question": content["sql_prompt"],
                        "schema": content["sql_context"],
                        "SQL": content["sql"],
                        "context": content["sql_context"],
                        "task_type": content["sql_task_type"],
                        "complexity": content["sql_complexity"],
                        "db_id": content["domain"],
                        "db_path": None,
                    }
                    dataset.append(blob_content)

            save_to_json(save_path=dataset_path / "train.json", json_object=dataset)

        super().__init__(
            split="train",
            dataset_path=dataset_path,
            database_folder_name=None,
            json_file_name="train.json",
        )

    def setup_dataset(
        self,
        filter_by: Optional[tuple] = None,
        num_rows: Optional[int] = None,
        num_fewshot: Optional[int] = None,
        model_name_or_path: Optional[str] = None,
        prompt_template: Optional[str] = BASE_TEXT2SQL_PROMPT,
    ):
        if filter_by:
            self.dataset = filter_options(data=self.dataset, filter_by=filter_by)

        if num_rows:
            self.dataset = self.dataset[:num_rows]

        self.dataset = GretelAIInstance(dataset=self.dataset).apply_prompt(
            num_fewshot=num_fewshot, prompt_template=prompt_template
        )

        return SupervisedDatasetForTraining(
            dataset=self.dataset,
            model_name_or_path=model_name_or_path,
            hf_token=self.hf_token,
        )



import logging
import os
import signal
import subprocess

from premsql.logger import setup_console_logger

logger = setup_console_logger("[BACKEND-UTILS]")


def stop_server_on_port(port: int):
    try:
        result = subprocess.run(
            ["lsof", "-ti", f":{port}"], capture_output=True, text=True
        )
        if result.returncode == 0:
            pid = int(result.stdout.strip())
            os.kill(pid, signal.SIGTERM)
            logger.info(f"Server running on port {port} (PID {pid}) stopped.")
        else:
            logger.info(f"No server found running on port {port}")
    except subprocess.CalledProcessError:
        logger.info(f"No server found running on port {port}")
    except ProcessLookupError:
        logger.info(f"Process on port {port} no longer exists")



from premsql.agents.baseline.workers import (
    BaseLineAnalyserWorker,
    BaseLineFollowupWorker,
    BaseLinePlotWorker,
    BaseLineText2SQLWorker,
)

__all__ = [
    "BaseLineAnalyserWorker",
    "BaseLineFollowupWorker",
    "BaseLinePlotWorker",
    "BaseLineText2SQLWorker",
]



from premsql.executors.from_langchain import ExecutorUsingLangChain
from premsql.executors.from_sqlite import SQLiteExecutor, OptimizedSQLiteExecutor

__all__ = ["ExecutorUsingLangChain", "SQLiteExecutor", "OptimizedSQLiteExecutor"]



import json
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import sqlparse
from tqdm.auto import tqdm
from platformdirs import user_cache_dir

from premsql.evaluator.base import BaseExecutor
from premsql.logger import setup_console_logger
from premsql.prompts import ERROR_HANDLING_PROMPT

logger = setup_console_logger(name="[GENERATOR]")


class Text2SQLGeneratorBase(ABC):
    def __init__(
        self, experiment_name: str, type: str, experiment_folder: Optional[str] = None
    ):
        self.experiment_folder = (
            Path(experiment_folder)
            if experiment_folder is not None
            else Path(user_cache_dir()) / "premsql" / "experiments"
        )

        self.experiment_path = Path(self.experiment_folder) / type / experiment_name
        if not self.experiment_path.exists():
            self.experiment_path.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created new experiment folder: {self.experiment_path}")
        else:
            logger.info(f"Experiment folder found in: {self.experiment_path}")

        self.client = self.load_client
        self.tokenizer = self.load_tokenizer

    @property
    @abstractmethod
    def load_client(self):
        return NotImplementedError

    @property
    @abstractmethod
    def load_tokenizer(self):
        return NotImplementedError

    @property
    @abstractmethod
    def model_name_or_path(self):
        pass

    @abstractmethod
    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs,
    ) -> str:
        raise NotImplementedError

    def execution_guided_decoding(
        self,
        data_blob: dict,
        executor: BaseExecutor,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        max_retries: Optional[int] = 5,
        postprocess: Optional[bool] = True,
        **kwargs,
    ):
        error_already_found = False
        for _ in range(max_retries):
            sql = self.generate(
                data_blob=data_blob,
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                postprocess=postprocess,
                **kwargs,
            )
            error = executor.execute_sql(sql=sql, dsn_or_db_path=data_blob["db_path"])[
                "error"
            ]
            if not error:
                return sql

            if not error_already_found:
                prompt = data_blob["prompt"].split("# SQL:")[0].strip()
                error_prompt = ERROR_HANDLING_PROMPT.format(
                    existing_prompt=prompt, sql=sql, error_msg=error
                )
                data_blob["prompt"] = error_prompt
                error_already_found = True
        return sql

    def postprocess(self, output_string: str):
        sql_start_keywords = [
            r"\bSELECT\b",
            r"\bINSERT\b",
            r"\bUPDATE\b",
            r"\bDELETE\b",
            r"\bWITH\b",
        ]

        sql_start_pattern = re.compile("|".join(sql_start_keywords), re.IGNORECASE)
        match = sql_start_pattern.search(output_string)
        if match:
            start_pos = match.start()
            sql_statement = output_string[start_pos:]
        else:
            sql_statement = output_string

        return sqlparse.format(sql_statement.split("# SQL:")[-1].strip())

    def load_results_from_folder(self):
        item_names = [item.name for item in self.experiment_path.iterdir()]

        if self.experiment_path.exists() and "predict.json" in item_names:
            return json.load(open(self.experiment_path / "predict.json", "r"))
        return None

    def generate_and_save_results(
        self,
        dataset: list[dict],
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        force: Optional[bool] = False,
        postprocess: Optional[bool] = False,
        executor: Optional[BaseExecutor] = None,
        max_retries: Optional[int] = 5,
        **kwargs,
    ) -> dict:

        existing_response = self.load_results_from_folder()
        if existing_response is not None and force == False:
            logger.info("Already results found")
            return existing_response

        to_dump = []
        for content in tqdm(dataset, total=len(dataset), desc="Generating result ..."):
            sql = (
                self.execution_guided_decoding(
                    data_blob=content,
                    executor=executor,
                    temperature=temperature,
                    postprocess=postprocess,
                    max_new_tokens=max_new_tokens,
                    max_retries=max_retries,
                    **kwargs,
                )
                if executor is not None
                else self.generate(
                    data_blob=content,
                    temperature=temperature,
                    max_new_tokens=max_new_tokens,
                    postprocess=postprocess,
                    **kwargs,
                )
            )

            to_dump.append({**content, "generated": sql})

        json.dump(to_dump, open(self.experiment_path / "predict.json", "w"), indent=4)
        logger.info(f"All responses are written to: {self.experiment_path}")
        return to_dump



from pathlib import Path
from typing import Optional, Union

from huggingface_hub import snapshot_download

from premsql.datasets.base import Text2SQLBaseDataset
from premsql.logger import setup_console_logger

logger = setup_console_logger("[BIRD-DATASET]")


class BirdDataset(Text2SQLBaseDataset):
    def __init__(
        self,
        split: str,
        dataset_folder: Optional[Union[str, Path]] = "./data",
        hf_token: Optional[str] = None,
        force_download: Optional[bool] = False,
        **kwargs
    ):
        dataset_folder = Path(dataset_folder)
        bird_folder = dataset_folder / "bird"
        if not bird_folder.exists() or force_download:
            bird_folder.mkdir(parents=True, exist_ok=True)

            # Download it from hf hub
            snapshot_download(
                repo_id="premai-io/birdbench",
                repo_type="dataset",
                local_dir=dataset_folder / "bird",
                force_download=force_download,
            )

        dataset_path = bird_folder / split

        database_folder_name = kwargs.get("database_folder_name", None) or (
            "train_databases" if split == "train" else "dev_databases"
        )
        json_file_name = kwargs.get("json_file_name", None) or (
            "train.json" if split == "train" else "validation.json"
        )

        super().__init__(
            split=split,
            dataset_path=dataset_path,
            database_folder_name=database_folder_name,
            json_file_name=json_file_name,
            hf_token=hf_token,
        )
        logger.info("Loaded Bird Dataset")

    def setup_dataset(
        self,
        filter_by: tuple | None = None,
        num_rows: int | None = None,
        num_fewshot: int | None = None,
        model_name_or_path: str | None = None,
        prompt_template: str | None = None,
        tokenize: bool | None = False 
    ):
        logger.info("Setting up Bird Dataset")
        return super().setup_dataset(
            filter_by=filter_by,
            num_rows=num_rows,
            model_name_or_path=model_name_or_path,
            tokenize=tokenize,
            prompt_template=prompt_template,
            num_fewshot=num_fewshot
        )


from premsql.playground.backend.backend_client import BackendAPIClient
from premsql.playground.inference_server.api_client import InferenceServerAPIClient
from premsql.playground.inference_server.service import AgentServer

__all__ = ["AgentServer", "InferenceServerAPIClient", "BackendAPIClient"]


import pandas as pd 
import streamlit as st
from premsql.playground.backend.backend_client import BackendAPIClient
from premsql.playground.inference_server.api_client import InferenceServerAPIClient
from premsql.playground.backend.api.pydantic_models import CompletionCreationRequest
from premsql.playground.frontend.components.streamlit_plot import StreamlitPlotTool
from premsql.agents.memory import AgentInteractionMemory
from premsql.agents.utils import convert_exit_output_to_agent_output
from premsql.agents.models import ExitWorkerOutput, AgentOutput
from premsql.logger import setup_console_logger

logger = setup_console_logger("FRONTEND-CHAT")

class ChatComponent:
    def __init__(self) -> None:
        self.backend_client = BackendAPIClient()
        self.inference_client = InferenceServerAPIClient()
        self.plotter = StreamlitPlotTool()
    
    def _streamlit_chat_output(self, message: AgentOutput | ExitWorkerOutput):
        if isinstance(message, ExitWorkerOutput):
            message = convert_exit_output_to_agent_output(exit_output=message)

        if message.output_dataframe:
            try:
                df = message.output_dataframe
                df = pd.DataFrame(df["data"], columns=df["columns"])
                if message.plot_config is None:
                    st.dataframe(df)
            except Exception as e:
                st.error(f"Error: {e}")

        if message.analysis:
            st.markdown(message.analysis)
        if message.plot_config:
            df = message.input_dataframe
            if df:
                self.plotter.run(
                    data=pd.DataFrame(df["data"], columns=df["columns"]),
                    plot_config=message.plot_config
                )
        if message.followup_suggestion:
            st.warning(message.followup_suggestion)
        with st.expander(label="Reasoning"):
            if message.sql_string:
                st.code(message.sql_string) 
            if message.reasoning:
                st.markdown(message.reasoning)
            if message.plot_config:
                st.json(message.plot_config)
            if message.error_from_pipeline:
                st.error(message.error_from_pipeline)


    def render_chat_env(self, session_name: str) -> None:
        session_info = self.backend_client.get_session(
            session_name=session_name
        )
        if session_info.status_code == 500:
            st.error(f"Failed to render chat History for session: {session_name}")

        session = session_info.sessions[0]
        session_db_path = session.session_db_path
        base_url = session.base_url
        # TODO: Need to understand how can I start the session

        history = AgentInteractionMemory(
            session_name=session_name, db_path=session_db_path
        )

        messages = history.generate_messages_from_session(session_name=session_name, server_mode=True)
        if not messages:
            st.warning("No chat history available for this session.")
        else:
            for message in messages:
                with st.chat_message("user"): st.markdown(message.question)
                with st.chat_message("assistant"):
                    self._streamlit_chat_output(message=message)
                        
        
        base_url = f"http://{base_url}" if not base_url.startswith("http://") else base_url
        is_session_online_status = self.inference_client.is_online(base_url=base_url)
        if is_session_online_status != 200:
            st.divider()
            st.warning(f"Session ended. Restart Agent Server to start the session at: {base_url}")
        
        else:
            if prompt := st.chat_input("What is your question?"):
                with st.chat_message("user"):
                    st.markdown(prompt)

                with st.chat_message("assistant"):
                    with st.spinner("Thinking..."):
                        response = self.backend_client.create_completion(
                            CompletionCreationRequest(
                                session_name=session_name,
                                question=prompt
                            )
                        )
                        if response.status_code == 200:
                            self._streamlit_chat_output(
                                message=history.get_by_message_id(message_id=response.message_id)
                            )
                        else:
                            st.error("Something went wrong. Try again")


                
        




import streamlit as st
from premsql.playground.frontend.components.chat import ChatComponent 
from premsql.playground.frontend.components.session import SessionComponent
from premsql.playground.frontend.components.uploader import UploadComponent

st.set_page_config(page_title="PremSQL Playground", page_icon="🔍", layout="wide")

def render_main_view():
    session_component = SessionComponent()
    
    selected_session = session_component.render_list_sessions()
    session_creation = session_component.render_register_session()
    session_component.render_additional_links()

    if session_creation is not None:
        if session_creation.status_code == 200:
            new_session_name = session_creation.session_name 
            st.success(f"New session created: {new_session_name}")
            ChatComponent().render_chat_env(session_name=new_session_name)
    elif selected_session is not None:
        ChatComponent().render_chat_env(session_name=selected_session)
    
    session_component.render_delete_session_view()

def main():
    _, col2, _ = st.sidebar.columns([1, 2, 1])
    with col2:
        st.image(
            "https://static.premai.io/logo.svg",
            use_container_width=True,
            width=150,
            clamp=True,
        )
        st.header("PremSQL Playground")
    st.title("PremSQL Playground")
    
    # Add navigation
    selected_page = st.sidebar.selectbox("Navigation", ["Chat", "Upload csvs or use Kaggle"])
    
    if selected_page == "Chat":
        st.write("Welcome to the PremSQL Playground. Select or create a session to get started.")
        render_main_view()
    else:
        st.write(
            "You can either upload multiple csv files or enter a valid Kaggle ID. "
            "This will migrate all the csvs into a SQLite Database. You can then "
            "use them for natural language powered analysis using PremSQL."
        )
        UploadComponent.render_kaggle_view()
        UploadComponent.render_csv_upload_view()

if __name__ == "__main__":
    main()


import random
import streamlit as st
from typing import Tuple, Optional
from pathlib import Path
from premsql.playground.frontend.utils import (
    download_from_kaggle,
    migrate_from_csv_to_sqlite,
    _is_valid_kaggle_id,
    migrate_local_csvs_to_sqlite
)

COMMON = """
db_connection_uri = "sqlite:///{db_path}"
baseline = BaseLineAgent(
    session_name="{session_name}",                # An unique session name must be put
    db_connection_uri=db_connection_uri,        # DB which needs to connect for Text to SQL 
    specialized_model1=text2sql_model,          # This referes to the Text to SQL model
    specialized_model2=analyser_plotter_model,  # This refers to any model other than Text to SQL
    executor=ExecutorUsingLangChain(),          # Which DB executor to use
    auto_filter_tables=False,                   # Whether to filter tables before Text to SQL
    plot_tool=SimpleMatplotlibTool()            # Matplotlib Tool which will be used by plotter worker
)

agent_server = AgentServer(agent=baseline, port={port})
agent_server.launch()
"""

STARTER_CODE_FILE_MLX = """
from premsql.playground import AgentServer
from premsql.agents import BaseLineAgent
from premsql.generators import Text2SQLGeneratorMLX
from premsql.executors import ExecutorUsingLangChain
from premsql.agents.tools import SimpleMatplotlibTool
text2sql_model = Text2SQLGeneratorMLX(
    model_name_or_path="premai-io/prem-1B-SQL", experiment_name="text2sql_model", type="test"
)

analyser_plotter_model = Text2SQLGeneratorMLX(
    model_name_or_path="meta-llama/Llama-3.2-1B-Instruct", experiment_name="analyser_model", type="test",
)
"""


STARTER_CODE_FILE_OLLAMA = """
from premsql.playground import AgentServer
from premsql.agents import BaseLineAgent
from premsql.generators import Text2SQLGeneratorOllama
from premsql.agents.tools import SimpleMatplotlibTool
from premsql.executors import ExecutorUsingLangChain

text2sql_model = Text2SQLGeneratorOllama(
    model_name="anindya/prem1b-sql-ollama-fp116",
    experiment_name="ollama",
    type="test"
)

analyser_plotter_model = Text2SQLGeneratorOllama(
    model_name="llama3.2:1b",
    experiment_name="ollama",
    type="test"
)
"""


STARTER_CODE_FILE_HF = """
from premsql.playground import AgentServer
from premsql.agents import BaseLineAgent
from premsql.generators import Text2SQLGeneratorHF
from premsql.executors import ExecutorUsingLangChain
from premsql.agents.tools import SimpleMatplotlibTool

text2sql_model = Text2SQLGeneratorHF(
    model_name_or_path="premai-io/prem-1B-SQL", experiment_name="text2sql_model", type="test"
)

analyser_plotter_model = Text2SQLGeneratorHF(
    model_name_or_path="meta-llama/Llama-3.2-1B-Instruct", experiment_name="analyser_model", type="test",
)
"""

STARTER_CODE_FILE_PREMAI = """
import os
from dotenv import load_dotenv
from premsql.playground import AgentServer
from premsql.agents import BaseLineAgent
from premsql.generators import Text2SQLGeneratorPremAI
from premsql.executors import ExecutorUsingLangChain
from premsql.agents.tools import SimpleMatplotlibTool

load_dotenv()

text2sql_model = Text2SQLGeneratorPremAI(
    model_name="gpt-4o", experiment_name="text2sql_model", type="test",
    premai_api_key=os.environ.get("PREMAI_API_KEY"),
    project_id=os.environ.get("PREMAI_PROJECT_ID")
)

analyser_plotter_model = Text2SQLGeneratorPremAI(
    model_name="gpt-4o", experiment_name="analyser_plotter_model", type="test",
    premai_api_key=os.environ.get("PREMAI_API_KEY"),
    project_id=os.environ.get("PREMAI_PROJECT_ID")
)
"""

STARTER_CODE_FILE_OPENAI = """
import os
from dotenv import load_dotenv
from premsql.playground import AgentServer
from premsql.agents import BaseLineAgent
from premsql.generators import Text2SQLGeneratorOpenAI
from premsql.executors import ExecutorUsingLangChain
from premsql.agents.tools import SimpleMatplotlibTool

load_dotenv()

text2sql_model = Text2SQLGeneratorOpenAI(
    model_name="gpt-4o", experiment_name="text2sql_model", type="test",
    openai_api_key=os.environ.get("OPENAI_API_KEY")
)

analyser_plotter_model = Text2SQLGeneratorOpenAI(
    model_name="gpt-4o", experiment_name="analyser_and_plotter_model", type="test",
    openai_api_key=os.environ.get("OPENAI_API_KEY")
)
"""

def render_starter_code(session_name, db_path):
    with st.expander(label="Start Locally with MLX", expanded=True):
        st.warning("Ensure you have mlx installed and using inside mac device.")
        st.info("Python PyPI: pip install mlx mlx-lm")

        code = (STARTER_CODE_FILE_MLX + COMMON).format(
            session_name=session_name, 
            db_path=db_path,
            port=random.choice(range(7000, 9000))
        )
        st.code(code, language="python")
    
    with st.expander(label="Start Locally with HuggingFace"):
        st.warning("Ensure you have torch, transformers installed inside your device.")
        st.info("Python PyPI: pip install torch transformers")

        code = (STARTER_CODE_FILE_HF + COMMON).format(
            session_name=session_name, 
            db_path=db_path,
            port=random.choice(range(7000, 9000))
        )
        st.code(code, language="python")

    with st.expander(label="Start Locally with Ollama"):
        st.warning("Ensure you have ollama installed inside your device.")
        st.info("Python PyPI: pip install ollama")
        st.info("Install Ollama: curl -fsSL https://ollama.com/install.sh | sh")

        code = (STARTER_CODE_FILE_OLLAMA + COMMON).format(
            session_name=session_name, 
            db_path=db_path,
            port=random.choice(range(7000, 9000))
        )
        st.code(code, language="python")

    with st.expander(label="Start with PremAI"):
        code = (STARTER_CODE_FILE_PREMAI + COMMON).format(
            session_name=session_name, 
            db_path=db_path,
            port=random.choice(range(7000, 9000))
        )
        st.code(code, language="python")

    with st.expander(label="Start with OpenAI"):
        code = (STARTER_CODE_FILE_OPENAI + COMMON).format(
            session_name=session_name, 
            db_path=db_path,
            port=random.choice(range(7000, 9000))
        )
        st.code(code, language="python")


class UploadComponent:
    @staticmethod
    def render_kaggle_view() -> Tuple[Optional[str], Optional[Path]]:
        session_name = None
        sqlite_db_path = None
        
        with st.sidebar:
            with st.expander(label="Upload From Kaggle"):
                with st.form(key="kaggle", clear_on_submit=True):
                    session_name = st.text_input(label="Enter session name")
                    kaggle_id = st.text_input(label="Enter kaggle id")
                    submit = st.form_submit_button(label="Submit")
                    
                    if submit:
                        if not session_name:
                            st.error("Please enter a session name")
                        
                        if not _is_valid_kaggle_id(kaggle_id):
                            st.error("Invalid Kaggle Id")
                        
                        try:
                            with st.spinner(text="Downloading from Kaggle"):
                                path = download_from_kaggle(kaggle_dataset_id=kaggle_id)
                            
                            sqlite_db_path = migrate_from_csv_to_sqlite(
                                folder_containing_csvs=path, 
                                session_name=session_name
                            )
                            st.success("Files downloaded and processed successfully!")
                        except Exception as e:
                            st.error(f"Error processing files: {str(e)}")
        if session_name and sqlite_db_path:
            render_starter_code(
                session_name=session_name, db_path=sqlite_db_path
            )
                    

    @staticmethod
    def render_csv_upload_view() -> Tuple[Optional[str], Optional[Path]]:
        with st.sidebar:
            with st.expander(label="Upload CSV Files"):
                with st.form(key="csv_upload", clear_on_submit=True):
                    session_name = st.text_input(label="Enter session name")
                    uploaded_files = st.file_uploader(
                        label="Upload CSV files",
                        type="csv",
                        accept_multiple_files=True
                    )
                    submit = st.form_submit_button(label="Submit")
                    
                    if submit:
                        if not session_name:
                            st.error("Please enter a session name")
                        
                        if not uploaded_files:
                            st.error("Please upload at least one CSV file")
                        
                        try:
                            with st.spinner(text="Processing CSV files"):
                                sqlite_db_path = migrate_local_csvs_to_sqlite(
                                    uploaded_files=uploaded_files,
                                    session_name=session_name
                                )
                            st.success("Files uploaded and processed successfully!")
                            
                        except Exception as e:
                            st.error(f"Error processing files: {str(e)}")
        if session_name and sqlite_db_path:
            render_starter_code(
                session_name=session_name, db_path=sqlite_db_path
            )
                    
    

    



import os
from typing import Optional, Union

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger

logger = setup_console_logger(name="[HF-GENERATOR]")

try:
    import torch
    import transformers
except ImportError:
    logger.warn("Ensure torch and transformers are installed.")
    logger.warn("Install them by: pip install torch transformers")

class Text2SQLGeneratorHF(Text2SQLGeneratorBase):
    def __init__(
        self,
        model_or_name_or_path: Union[str, "transformers.PreTrainedModel"],
        experiment_name: str,
        type: str,
        experiment_folder: Optional[str] = None,
        hf_token: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs
    ):
        self.hf_api_key = os.environ.get("HF_TOKEN") or hf_token
        self._kwargs = kwargs
        self.device = (
            device
            if device is not None
            else ("cuda:0" if torch.cuda.is_available() else "cpu")
        )
        self.model_or_name_or_path = model_or_name_or_path
        super().__init__(
            experiment_name=experiment_name,
            experiment_folder=experiment_folder,
            type=type,
        )

    @property
    def load_client(self) -> "transformers.PreTrainedModel":
        if isinstance(self.model_or_name_or_path, str):
            return transformers.AutoModelForCausalLM.from_pretrained(
                pretrained_model_name_or_path=self.model_or_name_or_path,
                token=self.hf_api_key,
                **{
                    "device_map": self.device,
                    "torch_dtype": torch.float16,
                    **self._kwargs,
                }
            )
        return self.model_or_name_or_path

    @property
    def load_tokenizer(self) -> "transformers.PreTrainedTokenizer":
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=self.client.config.name_or_path,
            token=self.hf_api_key,
            padding_side="right",
        )
        tokenizer.pad_token = tokenizer.eos_token
        return tokenizer

    @property
    def model_name_or_path(self):
        return self.model_or_name_or_path

    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs
    ) -> str:

        prompt = data_blob["prompt"]
        input_ids = self.tokenizer.encode(
            text=prompt,
            return_tensors="pt",
            padding="longest",
            max_length=self.tokenizer.model_max_length,
            truncation=False,
        ).to(self.device)

        do_sample = False if temperature == 0.0 else True
        generation_config = transformers.GenerationConfig(
            **{**kwargs, "temperature": temperature, "max_new_tokens": max_new_tokens}
        )
        output_tokens = (
            self.client.generate(
                input_ids=input_ids,
                do_sample=do_sample,
                generation_config=generation_config,
                pad_token_id=self.tokenizer.eos_token_id,
            )
            .detach()
            .tolist()[0]
        )
        output_tokens = (
            output_tokens[len(input_ids[0]) :]
            if len(output_tokens) > len(input_ids[0])
            else output_tokens
        )
        generated = self.tokenizer.decode(output_tokens, skip_special_tokens=True)
        return self.postprocess(output_string=generated) if postprocess else generated



from django.urls import path

from . import views

urlpatterns = [
    path("session/list/", views.list_sessions, name="list_sessions"),
    path("session/create", views.create_session, name="create_session"),
    path("session/<str:session_name>/", views.get_session, name="get_session"),
    path("session/<str:session_name>", views.delete_session, name="delete_session"),
    # Chat urls
    path("chat/completion", views.create_completion, name="completion"),
    path(
        "chat/history/<str:session_name>/", views.get_chat_history, name="chat_history"
    ),
]



from pathlib import Path
from typing import Optional, Union

from premsql.datasets.base import StandardDataset, Text2SQLBaseDataset
from premsql.datasets.real.bird import BirdDataset
from premsql.datasets.real.domains import DomainsDataset
from premsql.datasets.real.spider import SpiderUnifiedDataset
from premsql.datasets.synthetic.gretel import GretelAIDataset
from premsql.utils import get_accepted_filters


class Text2SQLDataset:
    def __init__(
        self,
        dataset_name: str,
        split: str,
        dataset_folder: Optional[Union[str, Path]] = "./data",
        hf_token: Optional[str] = None,
        force_download: Optional[bool] = False,
        **kwargs
    ):
        assert dataset_name in ["bird", "domains", "spider", "gretel"], ValueError(
            "Dataset should be one of bird, domains, spider, gretel"
        )
        dataset_mapping = {
            "bird": BirdDataset,
            "domains": DomainsDataset,
            "spider": SpiderUnifiedDataset,
            "gretel": GretelAIDataset,
        }
        self._text2sql_dataset: Text2SQLBaseDataset = dataset_mapping[dataset_name](
            split=split,
            dataset_folder=dataset_folder,
            hf_token=hf_token,
            force_download=force_download,
            **kwargs
        )

    @property
    def raw_dataset(self):
        return self._text2sql_dataset.dataset

    @property
    def filter_availables(self):
        return get_accepted_filters(data=self._text2sql_dataset.dataset)

    def setup_dataset(
        self,
        filter_by: tuple | None = None,
        num_rows: int | None = None,
        num_fewshot: int | None = None,
        model_name_or_path: str | None = None,
        prompt_template: str | None = None,
        tokenize: bool | None = False 
    ):
        return self._text2sql_dataset.setup_dataset(
            filter_by=filter_by,
            num_rows=num_rows,
            model_name_or_path=model_name_or_path,
            tokenize=tokenize,
            prompt_template=prompt_template,
            num_fewshot=num_fewshot
        )


__all__ = [
    "StandardDataset",
    "GretelAIDataset",
    "SpiderUnifiedDataset",
    "BirdDataset",
    "DomainsDataset",
    "Text2SQLDataset",
]


import traceback
from typing import Dict, Any
import pandas as pd
import streamlit as st 
from premsql.logger import setup_console_logger
from premsql.agents.tools.plot.base import BasePlotTool

logger = setup_console_logger("[STREAMLIT-TOOL]")

class StreamlitPlotTool(BasePlotTool):
    def __init__(self):
        self.plot_functions = {
            "area": self._area_plot,
            "bar": self._bar_plot,
            "scatter": self._scatter_plot,
            "histogram": self._histogram_plot,
            "line": self._line_plot,
        }

    def run(self, data: pd.DataFrame, plot_config: Dict[str, str]) -> Any:
        try:
            self._validate_config(data, plot_config)

            plot_type = plot_config["plot_type"]
            x = plot_config["x"]
            y = plot_config["y"]

            st.markdown(f"**{plot_type.capitalize()} Plot: {x} vs {y}**")
            return self.plot_functions[plot_type](data, x, y)
        except Exception as e:
            error_msg = f"Error creating plot: {str(e)}"
            stack_trace = traceback.format_exc()
            logger.error(f"{error_msg}\n{stack_trace}")
            logger.error(f"Error creating plot: {str(e)}")
            st.error(f"Error creating plot: {str(e)}")
            return None

    def _validate_config(self, df: pd.DataFrame, plot_config: Dict[str, str]) -> None:
        required_keys = ["plot_type", "x", "y"]
        missing_keys = [key for key in required_keys if key not in plot_config]
        if missing_keys:
            raise ValueError(f"Missing required keys in plot_config: {', '.join(missing_keys)}")

        for key in ["x", "y"]:
            if key not in plot_config:
                raise ValueError(f"'{key}' is missing from plot_config")
            if not isinstance(plot_config[key], str):
                raise TypeError(f"plot_config['{key}'] should be a string, but got {type(plot_config[key])}")

        if not isinstance(df, pd.DataFrame):
            raise TypeError(f"Expected df to be a pandas DataFrame, but got {type(df)}")

        if not hasattr(df, 'columns'):
            raise AttributeError(f"df does not have a 'columns' attribute. Type: {type(df)}")

        if plot_config["x"] not in df.columns:
            raise ValueError(f"Column '{plot_config['x']}' not found in DataFrame. Available columns: {', '.join(df.columns)}")

        if plot_config["y"] not in df.columns:
            raise ValueError(f"Column '{plot_config['y']}' not found in DataFrame. Available columns: {', '.join(df.columns)}")

        if plot_config["plot_type"] not in self.plot_functions:
            raise ValueError(f"Unsupported plot type: {plot_config['plot_type']}. Supported types: {', '.join(self.plot_functions.keys())}")

    def _area_plot(self, df: pd.DataFrame, x: str, y: str) -> Any:
        chart_data = df[[x, y]].set_index(x)
        return st.area_chart(chart_data)

    def _bar_plot(self, df: pd.DataFrame, x: str, y: str) -> Any:
        chart_data = df[[x, y]].set_index(x)
        return st.bar_chart(chart_data)

    def _scatter_plot(self, df: pd.DataFrame, x: str, y: str) -> Any:
        chart_data = df[[x, y]]
        return st.scatter_chart(chart_data, x=x, y=y)

    def _histogram_plot(self, df: pd.DataFrame, x: str, y: str) -> Any:
        # Streamlit doesn't have a built-in histogram function, so we'll use a bar chart
        hist_data = df[x].value_counts().sort_index()
        chart_data = pd.DataFrame({x: hist_data.index, 'count': hist_data.values})
        return st.bar_chart(chart_data.set_index(x))

    def _line_plot(self, df: pd.DataFrame, x: str, y: str) -> Any:
        chart_data = df[[x, y]].set_index(x)
        return st.line_chart(chart_data)
    
    def convert_plot_to_image(self, fig):
        pass 


from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests


class InferenceServerAPIError(Exception):
    pass


class InferenceServerAPIClient:
    def __init__(self, timeout: int = 600) -> None:
        self.headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
        }
        self.timeout = timeout

    def _make_request(
        self,
        base_url: str,
        method: str,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        url = urljoin(base_url.rstrip("/"), endpoint)
        try:
            response = requests.request(
                method, url, headers=self.headers, json=data, timeout=self.timeout
            )
            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            raise InferenceServerAPIError(f"API request failed: {str(e)}")
    
    def is_online(self, base_url: str) -> bool:
        endpoint = "/health"
        try:
            response = self._make_request(base_url, "GET", endpoint)
            return response.get("status_code")
        except Exception as e:
            return 500

    def post_completion(self, base_url: str, question: str) -> Dict[str, Any]:
        if not question.strip():
            raise ValueError("Question cannot be empty")
        endpoint = "/completion"
        data = {"question": question}
        return self._make_request(base_url, "POST", endpoint, data)

    def get_session_info(self, base_url: str) -> Dict[str, Any]:
        endpoint = "/session_info"
        return self._make_request(base_url, "GET", endpoint)

    def get_chat_history(self, base_url: str, message_id: int) -> Dict[str, Any]:
        if message_id < 1:
            raise ValueError("Message ID must be a positive integer")
        endpoint = f"/chat_history/{message_id}"
        return self._make_request(base_url, "GET", endpoint)

    def delete_session(self, base_url: str) -> Dict[str, Any]:
        endpoint = "/delete_session/"
        return self._make_request(base_url, "DELETE", endpoint)



import json
import os
import random
import re
import sqlite3
from collections import defaultdict
from pathlib import Path
from textwrap import dedent
from typing import Optional, Sequence, Union

from tqdm.auto import tqdm
from premsql.logger import setup_console_logger

logger = setup_console_logger(name="[UTILS]")

try:
    from transformers import PreTrainedTokenizer
except ImportError:
    logger.warn("Unable to use transformers. Install using: pip install transformers")


def convert_sqlite_path_to_dsn(path: str):
    sqlite3_pattern = r"^sqlite:\/\/\/.*"
    if re.match(sqlite3_pattern, path):
        return path
    return f"sqlite:///{os.path.abspath(path)}"


def convert_sqlite_dsn_to_path(dsn: str) -> str:
    sqlite3_pattern = r"^sqlite:\/\/\/(.*)"
    match = re.match(sqlite3_pattern, dsn)
    if match:
        return os.path.abspath(match.group(1))
    return dsn


def print_data(data: dict):
    if "prompt" in data:
        prompt = data["prompt"]
        data["prompt"] = prompt[:100] + "...." + prompt[-100:]

    elif "prompt" in data["raw"]:
        prompt = data["raw"]["prompt"]
        data["raw"]["prompt"] = prompt[:100] + "...." + prompt[-100:]

    else:
        raise ValueError("Prompt key not found in data")

    return data


def save_to_json(save_path: Union[str, Path], json_object: dict):
    try:
        save_path = Path(save_path) if isinstance(save_path, str) else save_path
        with open(save_path, "w") as json_file:
            json.dump(json_object, json_file, indent=4)
        logger.info(f"Saved JSON in: {save_path}")
    except Exception as e:
        logger.error(f"Unable to save JSON, Error: {e}")


def load_from_json(result_json_path: str) -> dict:
    try:
        with open(result_json_path, "r") as json_file:
            return json.load(json_file)
    except Exception as e:
        logger.error(f"Unable to load JSON, Error: {e}")


def sqlite_schema_prompt(db_path: str) -> str:
    schemas = {}
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()

    for table in tables:
        table_name = table[0]
        if table_name == "sqlite_sequence":
            continue
        cursor.execute(
            f"SELECT sql FROM sqlite_master WHERE type='table' AND name='{table_name}';"
        )
        create_table_sql = cursor.fetchone()
        if create_table_sql:
            schemas[table_name] = create_table_sql[0]
        else:
            schemas[table_name] = "Schema does not exist"

    schema_prompt = "\n".join(
        schemas[table[0]] for table in tables if table[0] != "sqlite_sequence"
    )
    return schema_prompt


def get_random_few_shot_prompts(dataset: list[dict], num_few_shot: int):
    assert "db_id" in dataset[0], ValueError(
        "db_id key should be present to use this function"
    )

    grouped_content = defaultdict(list)
    few_shot_prompts = {}
    template = dedent(
        """
    Question: {question}
    SQL: {sql}
    """
    )

    for content in dataset:
        grouped_content[content["db_id"]].append(content)

    for db_id, contents in grouped_content.items():
        num_few_shot = min(num_few_shot, len(contents))
        random_sample = random.sample(contents, num_few_shot)

        few_shot_prompt = "".join(
            template.format(question=element["question"], sql=element["SQL"])
            for element in random_sample
        )
        few_shot_prompts[db_id] = few_shot_prompt
    return few_shot_prompts


def get_accepted_filters(data: list[dict]) -> Sequence[str]:
    key_num_mapping = {}
    for key in data[0].keys():
        key_num_mapping[key] = len(set([content[key] for content in data]))

    accepted_keys = []
    for key, num in key_num_mapping.items():
        if num < len(data) * 0.5 and key != "db_path":
            accepted_keys.append(key)
    return accepted_keys


def filter_options(
    data: list[dict], filter_by: tuple, accepted_keys: Optional[Sequence[str]] = None
):
    filter_key, filter_value = filter_by
    accepted_keys = (
        get_accepted_filters(data=data) if accepted_keys is None else accepted_keys
    )

    assert filter_key in accepted_keys, ValueError(
        f"Filtering is supported for keys: `{''.join(accepted_keys)}`"
    )
    for key in accepted_keys:
        if filter_key == key:
            accepted_values = set([content[key] for content in data])
            assert filter_value in accepted_values, ValueError(
                f"Available values for key: {key} are: {', '.join(accepted_values)}"
            )

    filtered_data = [content for content in data if content[filter_key] == filter_value]
    return filtered_data


def tokenize_fn(strings: Sequence[str], tokenizer: "PreTrainedTokenizer") -> dict:
    """Tokenizes a list of string"""
    tokenized_list = [
        tokenizer(
            text=text,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=False,
        )
        for text in tqdm(strings, total=len(strings), desc="Tokenizing")
    ]
    input_ids = labels = [tokenized.input_ids[0] for tokenized in tokenized_list]
    input_ids_lens = label_ids_lens = [
        tokenized.input_ids.ne(tokenizer.pad_token_id).sum().item()
        for tokenized in tokenized_list
    ]
    return dict(
        input_ids=input_ids,
        labels=labels,
        input_ids_lens=input_ids_lens,
        label_ids_lens=label_ids_lens,
    )



from typing import Optional

import pandas as pd

from premsql.logger import setup_console_logger
from premsql.agents.base import RouterWorkerBase, RouterWorkerOutput
from premsql.agents.utils import convert_df_to_dict

logger = setup_console_logger("[BASELINE-ROUTER]")


class SimpleRouterWorker(RouterWorkerBase):
    def run(
        self, question: str, input_dataframe: Optional[pd.DataFrame]
    ) -> RouterWorkerOutput:
        if question.startswith("/query"):
            route_to = "query"
        elif question.startswith("/analyse"):
            route_to = "analyse"
        elif question.startswith("/plot"):
            route_to = "plot"
        else:
            route_to = "followup"
        logger.info(f"Routing to: {route_to}")
        question = (
            question.split(f"/{route_to}")[1] if route_to != "followup" else question
        )

        return RouterWorkerOutput(
            question=question,
            route_to=route_to,
            input_dataframe=(
                convert_df_to_dict(df=input_dataframe) if input_dataframe else None
            ),
            decision_reasoning="Simple routing based on question prefix",
            additional_input={},
            error_from_model=None,
        )



import os
from typing import Optional

from premsql.generators.base import Text2SQLGeneratorBase

try:
    from openai import OpenAI
except ImportError:
    raise ImportError("Module openai is not installed")


class Text2SQLGeneratorOpenAI(Text2SQLGeneratorBase):
    def __init__(
        self,
        model_name: str,
        experiment_name: str,
        type: str,
        experiment_folder: Optional[str] = None,
        openai_api_key: Optional[str] = None,
    ):
        self._api_key = openai_api_key or os.environ.get("OPENAI_API_KEY")
        self.model_name = model_name
        super().__init__(
            experiment_folder=experiment_folder,
            experiment_name=experiment_name,
            type=type,
        )

    @property
    def load_client(self):
        client = OpenAI(api_key=self._api_key)
        return client

    @property
    def load_tokenizer(self):
        pass

    @property
    def model_name_or_path(self):
        return self.model_name

    def generate(
        self,
        data_blob: dict,
        temperature: Optional[float] = 0.0,
        max_new_tokens: Optional[int] = 256,
        postprocess: Optional[bool] = True,
        **kwargs
    ) -> str:
        prompt = data_blob["prompt"]
        max_tokens = max_new_tokens
        generation_config = {
            **kwargs,
            **{"temperature": temperature, "max_tokens": max_tokens},
        }
        completion = (
            self.client.chat.completions.create(
                model=self.model_name,
                messages=[{"role": "user", "content": prompt}],
                **generation_config
            )
            .choices[0]
            .message.content
        )
        return self.postprocess(output_string=completion) if postprocess else completion



from datetime import datetime
from typing import List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from premsql.agents.models import AgentOutput

# All the Session Models


class SessionCreationRequest(BaseModel):
    base_url: str = Field(...)
    model_config = ConfigDict(extra="forbid")


class SessionCreationResponse(BaseModel):
    status_code: Literal[200, 500] = Field(...)
    status: Literal["success", "error"] = Field(...)

    session_id: Optional[int] = None
    session_name: Optional[str] = None
    db_connection_uri: str = Field(None)
    session_db_path: str = Field(None)
    created_at: Optional[datetime] = None
    error_message: Optional[str] = None


class SessionSummary(BaseModel):
    session_id: int
    session_name: str
    created_at: datetime
    base_url: str
    db_connection_uri: str
    session_db_path: str

    model_config = ConfigDict(from_attributes=True)


class SessionListResponse(BaseModel):
    status_code: Literal[200, 500]
    status: Literal["success", "error"]
    sessions: Optional[List[SessionSummary]] = None
    total_count: Optional[int] = None
    page: Optional[int] = None
    page_size: Optional[int] = None
    error_message: Optional[str] = None


class SessionDeleteResponse(BaseModel):
    session_name: str
    status_code: Literal[200, 404, 500]
    status: Literal["success", "error"]
    error_message: Optional[str] = None


# All the chat message models


class CompletionCreationRequest(BaseModel):
    session_name: str
    question: str


class CompletionCreationResponse(BaseModel):
    status_code: Literal[200, 500]
    status: Literal["success", "error"]
    message_id: Optional[int] = None
    session_name: Optional[str] = None
    created_at: Optional[datetime] = None
    message: Optional[AgentOutput] = None
    question: Optional[str] = None
    error_message: Optional[str] = None


class CompletionSummary(BaseModel):
    message_id: int
    session_name: str
    base_url: str
    created_at: datetime
    question: Optional[str] = None

    model_config = ConfigDict(from_attributes=True)


class CompletionListResponse(BaseModel):
    status_code: Literal[200, 500]
    status: Literal["success", "error"]
    completions: Optional[List[CompletionSummary]] = None
    total_count: Optional[int] = None
    error_message: Optional[str] = None



from django.db import models


class Session(models.Model):
    session_id = models.AutoField(primary_key=True)
    db_connection_uri = models.URLField()
    session_name = models.CharField(max_length=255, unique=True)
    created_at = models.DateTimeField(auto_now_add=True)
    base_url = models.URLField()
    session_db_path = models.CharField(max_length=255)

    class Meta:
        ordering = ["created_at"]


class Completions(models.Model):
    chat_id = models.AutoField(primary_key=True)
    message_id = models.IntegerField(blank=True, null=True)
    session = models.ForeignKey(
        Session, on_delete=models.CASCADE, related_name="messages"
    )
    session_name = models.CharField(max_length=255)
    created_at = models.DateTimeField()
    question = models.TextField(blank=True, null=True)

    class Meta:
        ordering = ["-created_at"]
        verbose_name_plural = "Completions"



from dataclasses import dataclass
from typing import Sequence
from premsql.logger import setup_console_logger

logger = setup_console_logger("[DATASET-COLLATOR]")

try:
    import torch
    import transformers
except ImportError:
    logger.warn("Ensure torch and transformers are installed.")
    logger.warn("Install them by: pip install torch transformers")

@dataclass
class DataCollatorForSupervisedDataset:
    tokenizer: "transformers.PreTrainedTokenizer"

    def __call__(self, instances: Sequence[dict]) -> dict[str, torch.Tensor]:
        input_ids, labels = tuple(
            [instance[key] for instance in instances] for key in ("input_ids", "labels")
        )
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids,
            batch_first=True,
            padding_value=self.tokenizer.pad_token_id,
        )
        labels = torch.nn.utils.rnn.pad_sequence(
            labels, batch_first=True, padding_value=-100
        )
        return dict(
            input_ids=input_ids,
            labels=labels,
            attention_mask=input_ids.ne(self.tokenizer.pad_token_id),
        )



import base64
import io
from abc import ABC, abstractmethod

import pandas as pd
from PIL import Image


class BasePlotTool(ABC):
    @abstractmethod
    def run(self, data: pd.DataFrame, plot_config: dict):
        raise NotImplementedError()

    @abstractmethod
    def convert_plot_to_image(self, fig):
        raise NotImplementedError

    def convert_image_to_base64(self, image: Image.Image) -> str:
        buffered = io.BytesIO()
        image.save(buffered, format="PNG")
        return base64.b64encode(buffered.getvalue()).decode()

    def save_image(self, image: Image.Image, file_path: str, format: str = "PNG"):
        image.save(file_path, format=format)

    def plot_from_base64(self, output_base64: str):
        image_data = base64.b64decode(output_base64)
        return Image.open(io.BytesIO(image_data))



from typing import Optional, Sequence

import transformers

from premsql.datasets.base import Text2SQLBaseDataset
from premsql.datasets.collator import DataCollatorForSupervisedDataset
from premsql.evaluator.base import BaseExecutor
from premsql.logger import setup_console_logger
from premsql.tuner.callback import Text2SQLEvaluationCallback
from premsql.tuner.config import DefaultTrainingArguments

logger = setup_console_logger("[FULL-FINETUNE]")


class Text2SQLFullFinetuner:
    def __init__(
        self,
        model_name_or_path: str,
        experiment_name: str,
        hf_token: Optional[str] = None,
        **model_kwargs,
    ):
        self.model_name_or_path = model_name_or_path

        logger.warning("Setting up Pretrained-Model: " + str(model_name_or_path))
        self.model = transformers.AutoModelForCausalLM.from_pretrained(
            model_name_or_path, token=hf_token, **model_kwargs
        )
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(
            model_name_or_path, padding_size="right", token=hf_token
        )
        self.data_collator = DataCollatorForSupervisedDataset(tokenizer=self.tokenizer)

        self._hf_token = hf_token
        self.experiment_name = experiment_name

    def train(
        self,
        train_datasets: Sequence[Text2SQLBaseDataset],
        output_dir: str,
        num_train_epochs: int,
        per_device_train_batch_size: int,
        gradient_accumulation_steps: int,
        evaluation_dataset: Optional[Text2SQLBaseDataset] = None,
        eval_steps: Optional[int] = 500,
        executor: Optional[BaseExecutor] = None,
        filter_eval_results_by: Optional[tuple] = None,
        **training_arguments,
    ):
        self.training_arguments = DefaultTrainingArguments(
            output_dir=output_dir,
            num_train_epochs=num_train_epochs,
            per_device_train_batch_size=per_device_train_batch_size,
            gradient_accumulation_steps=gradient_accumulation_steps,
            **training_arguments,
        )

        data_module = dict(
            train_dataset=train_datasets,
            eval_dataset=None,
            data_collator=self.data_collator,
        )
        trainer = transformers.Trainer(
            model=self.model,
            tokenizer=self.tokenizer,
            args=self.training_arguments,
            **data_module,
        )

        if evaluation_dataset is not None and executor is not None:
            eval_callback = Text2SQLEvaluationCallback(
                trainer=trainer,
                trainer_args=self.training_arguments,
                eval_dataset=evaluation_dataset,
                experiment_name=self.experiment_name,
                model_or_name_or_id=self.model_name_or_path,
                eval_steps=eval_steps,
                executor=executor,
                filter_results_by=filter_eval_results_by,
                hf_token=self._hf_token,
            )
            trainer.add_callback(eval_callback)

        trainer.train()
        trainer.save_model(output_dir=self.training_arguments.output_dir)



from premsql.tuner.callback import Text2SQLEvaluationCallback
from premsql.tuner.config import (
    DefaultLoraConfig,
    DefaultPeftArguments,
    DefaultTrainingArguments,
)
from premsql.tuner.full import Text2SQLFullFinetuner
from premsql.tuner.peft import Text2SQLPeftTuner

__all__ = [
    "Text2SQLFullFinetuner",
    "Text2SQLPeftTuner",
    "DefaultLoraConfig",
    "DefaultPeftArguments",
    "Text2SQLEvaluationCallback",
]



import os
from typing import Optional

from premsql.datasets.base import Text2SQLBaseDataset
from premsql.evaluator.base import BaseExecutor, Text2SQLEvaluator
from premsql.generators.huggingface import Text2SQLGeneratorHF
from premsql.logger import setup_console_logger

logger = setup_console_logger("[EVALUATION-CALLBACK]")

try:
    from torch.utils.tensorboard import SummaryWriter
    from transformers import (
        Trainer,
        TrainerCallback,
        TrainerControl,
        TrainerState,
        TrainingArguments,
    )
except ImportError:
    logger.warn("Unable to import torch and transformers. Install: pip install torch transformers")


class Text2SQLEvaluationCallback(TrainerCallback):
    def __init__(
        self,
        trainer: Trainer,
        trainer_args: TrainingArguments,
        eval_dataset: Text2SQLBaseDataset,
        executor: BaseExecutor,
        experiment_name: str,
        model_or_name_or_id: str,
        eval_steps: int,
        hf_token: Optional[str] = None,
        filter_results_by: Optional[tuple] = None,
    ):
        self.trainer = trainer
        self.eval_steps = eval_steps
        self.experiment_name = experiment_name

        log_dir = trainer_args.logging_dir
        os.makedirs(log_dir, exist_ok=True)

        self.tb_writer = SummaryWriter(log_dir=log_dir)
        logger.info(f"TensorBoard log directory: {log_dir}")

        self.model_or_name_or_id = model_or_name_or_id
        self.hf_token = hf_token
        self.dataset = eval_dataset
        self.executor = executor
        self.filter_by = filter_results_by

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        if args.local_rank == 0 and state.global_step % self.eval_steps == 0:
            logger.info(f"Evaluating at step {state.global_step}")
            model = Text2SQLGeneratorHF(
                model_or_name_or_path=self.trainer.model,
                experiment_name=f"{self.experiment_name}_step_{state.global_step}",
                type="test",
                device="cuda:0",
            )
            responses = model.generate_and_save_results(
                dataset=self.dataset, temperature=0.1, max_new_tokens=256, force=True
            )
            evaluator = Text2SQLEvaluator(
                executor=self.executor, experiment_path=model.experiment_path
            )
            if self.filter_by:
                ex_score = evaluator.execute(
                    metric_name="accuracy",
                    model_responses=responses,
                    filter_by=self.filter_by[0],
                )
            else:
                ex_score = evaluator.execute(
                    metric_name="accuracy",
                    model_responses=responses,
                )
            logger.info(f"Execution Accuracy at step {state.global_step} | {ex_score}")

            # Log into tensorboard
            logger.info(f"Logging to TensorBoard: {ex_score}")
            for difficulty, score in ex_score.items():
                logger.info(f"Logging {difficulty}: {score}")
                self.tb_writer.add_scalar(
                    f"execution_accuracy/{difficulty}", score, state.global_step
                )
            self.tb_writer.flush()  # Force writing to disk

            state.log_history.append(
                {
                    "step": state.global_step,
                    "execution_accuracy": (
                        ex_score.get(self.filter_by[1])
                        if self.filter_by
                        else ex_score.get("overall")
                    ),
                    "selected_difficulty": (
                        self.filter_by[0] if self.filter_by else "overall"
                    ),
                }
            )
        return control

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self.tb_writer.close()
        logger.info("TensorBoard writer closed")



from premsql.agents.baseline.workers.analyser import BaseLineAnalyserWorker
from premsql.agents.baseline.workers.followup import BaseLineFollowupWorker
from premsql.agents.baseline.workers.plotter import BaseLinePlotWorker
from premsql.agents.baseline.workers.text2sql import BaseLineText2SQLWorker

__all__ = [
    "BaseLineText2SQLWorker",
    "BaseLineAnalyserWorker",
    "BaseLinePlotWorker",
    "BaseLineFollowupWorker",
]



import logging


def setup_console_logger(name, level=logging.INFO):
    """Function to setup a console logger."""
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(console_handler)

    return logger



import json

from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from premsql.logger import setup_console_logger
from premsql.playground.backend.api.pydantic_models import (
    CompletionCreationRequest,
    SessionCreationRequest,
    SessionListResponse,
    SessionSummary,
)
from premsql.playground.backend.api.serializers import (
    CompletionCreationRequestSerializer,
    CompletionCreationResponseSerializer,
    CompletionListResponseSerializer,
    SessionCreationRequestSerializer,
    SessionCreationResponseSerializer,
    SessionListResponseSerializer,
    SessionSummarySerializer,
)

from .services import CompletionService, SessionManageService

logger = setup_console_logger("[VIEWS]")


@swagger_auto_schema(
    method="post",
    request_body=SessionCreationRequestSerializer,
    responses={
        200: SessionCreationResponseSerializer,
        400: "Bad Request",
        500: SessionCreationResponseSerializer,
    },
)
@api_view(["POST"])
def create_session(request):
    try:
        session_request = SessionCreationRequest(**request.data)
        response = SessionManageService().create_session(request=session_request)
        return Response(response.model_dump())
    except json.JSONDecodeError:
        return Response(
            {"status": "error", "error_message": "Invalid JSON"},
            status=status.HTTP_400_BAD_REQUEST,
        )
    except Exception as e:
        return Response(
            {"status": "error", "error_message": str(e)},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )


@swagger_auto_schema(
    method="get",
    manual_parameters=[
        openapi.Parameter(
            "session_name",
            openapi.IN_PATH,
            description="Name of the session",
            type=openapi.TYPE_STRING,
        ),
    ],
    responses={
        200: SessionSummarySerializer,
        400: "Bad Request",
        500: SessionSummarySerializer,
    },
)
@api_view(["GET"])
def get_session(request, session_name):
    session = SessionManageService().get_session(session_name=session_name)
    if session:
        session_summary = SessionSummary.model_validate(session)
        response = SessionListResponse(
            status="success",
            status_code=200,
            sessions=[session_summary.model_dump()],
            total_count=1,
            page=1,
            page_size=1,
        )
    else:
        response = SessionListResponse(
            status="error",
            status_code=500,
            error_message="The requested session does not exist.",
        )
    return Response(
        response.model_dump(),
        status=status.HTTP_200_OK if session else status.HTTP_404_NOT_FOUND,
    )


@swagger_auto_schema(
    method="get",
    manual_parameters=[
        openapi.Parameter(
            "page",
            openapi.IN_QUERY,
            description="Page number",
            type=openapi.TYPE_INTEGER,
            default=1,
        ),
        openapi.Parameter(
            "page_size",
            openapi.IN_QUERY,
            description="Number of items per page",
            type=openapi.TYPE_INTEGER,
            default=20,
        ),
    ],
    responses={
        200: SessionListResponseSerializer,
        400: "Bad Request",
        500: SessionListResponseSerializer,
    },
)
@api_view(["GET"])
def list_sessions(request):
    page = int(request.query_params.get("page", 1))
    page_size = int(request.query_params.get("page_size", 20))
    response = SessionManageService().list_session(page=page, page_size=page_size)
    return Response(response.model_dump())


@swagger_auto_schema(
    method="delete",
    manual_parameters=[
        openapi.Parameter(
            "session_name",
            openapi.IN_PATH,
            description="Name of the session to delete",
            type=openapi.TYPE_STRING,
            required=True,
        ),
    ],
    responses={
        200: openapi.Response(
            "Session deleted successfully",
            schema=openapi.Schema(
                type=openapi.TYPE_OBJECT,
                properties={
                    "status": openapi.Schema(
                        type=openapi.TYPE_STRING, example="success"
                    ),
                    "message": openapi.Schema(type=openapi.TYPE_STRING),
                },
            ),
        ),
        404: "Not Found",
        500: "Internal Server Error",
    },
)
@api_view(["DELETE"])
def delete_session(request, session_name):
    try:
        result = SessionManageService().delete_session(session_name=session_name)
        return Response(result.model_dump(), status=result.status_code)
    except Exception as e:
        return Response(
            {"status": "error", "error_message": str(e)},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )


# Completion Views


@swagger_auto_schema(
    method="post",
    request_body=CompletionCreationRequestSerializer,
    responses={
        200: CompletionCreationResponseSerializer,
        400: "Bad Request",
        404: "Not Found",
        500: "Internal Server Error",
    },
)
@api_view(["POST"])
def create_completion(request):
    try:
        completion_request = CompletionCreationRequest(**request.data)
        response = CompletionService().completion(request=completion_request)
        return Response(
            response.model_dump(),
            status=(
                status.HTTP_200_OK
                if response.status == "success"
                else status.HTTP_500_INTERNAL_SERVER_ERROR
            ),
        )
    except ValidationError as e:
        return Response(
            {"status": "error", "error_message": str(e)},
            status=status.HTTP_400_BAD_REQUEST,
        )
    except Exception as e:
        return Response(
            {"status": "error", "error_message": str(e)},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )


@swagger_auto_schema(
    method="get",
    manual_parameters=[
        openapi.Parameter(
            "session_name",
            openapi.IN_PATH,
            description="Name of the session",
            type=openapi.TYPE_STRING,
            required=True,
        ),
        openapi.Parameter(
            "page",
            openapi.IN_QUERY,
            description="Page number",
            type=openapi.TYPE_INTEGER,
            default=1,
        ),
        openapi.Parameter(
            "page_size",
            openapi.IN_QUERY,
            description="Number of items per page",
            type=openapi.TYPE_INTEGER,
            default=20,
        ),
    ],
    responses={
        200: CompletionListResponseSerializer,
        400: "Bad Request",
        404: "Not Found",
        500: "Internal Server Error",
    },
)
@api_view(["GET"])
def get_chat_history(request, session_name):
    try:
        page = int(request.query_params.get("page", 1))
        page_size = int(request.query_params.get("page_size", 20))

        response = CompletionService().chat_history(
            session_name=session_name, page=page, page_size=page_size
        )

        return Response(
            response.model_dump(),
            status=(
                status.HTTP_200_OK
                if response.status == "success"
                else status.HTTP_404_NOT_FOUND
            ),
        )
    except ValueError:
        return Response(
            {"status": "error", "error_message": "Invalid page or page_size parameter"},
            status=status.HTTP_400_BAD_REQUEST,
        )
    except Exception as e:
        return Response(
            {"status": "error", "error_message": str(e)},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )



from dataclasses import dataclass, field
from typing import List, Optional
from premsql.logger import setup_console_logger

logger = setup_console_logger("[TUNER-CONFIG]")

try:
    from peft import LoraConfig, TaskType
    from transformers import TrainingArguments
except ImportError:
    logger.warn("Unable to find peft and transformers. Install: pip install peft transformers")


@dataclass
class DefaultTrainingArguments(TrainingArguments):
    output_dir: str
    num_train_epochs: int
    per_device_train_batch_size: int
    gradient_accumulation_steps: int

    load_best_model_at_end: Optional[bool] = field(default=True)
    gradient_checkpointing: Optional[bool] = field(default=True)
    evaluation_strategy: Optional[str] = field(default="no")

    cache_dir: Optional[str] = field(default=None)
    optim: str = field(default="adamw_hf")
    model_max_length: int = field(
        default=1024,
        metadata={
            "help": "Maximum sequence length. Sequences will be right padded (and possibly truncated)."
        },
    )
    max_seq_length: int = field(default=1024)
    ddp_find_unused_parameters: Optional[bool] = field(default=False)
    fp16: bool = field(default=False)
    bf16: bool = field(default=True)

    weight_decay: float = field(default=0.1)
    lr_scheduler_type: str = field(default="cosine")
    warmup_ratio: float = field(default=0.01)
    logging_steps: int = field(default=10)
    save_strategy: str = field(default="steps")
    save_steps: int = field(default=200)
    save_total_limit: int = field(default=3)
    auto_find_batch_size: Optional[bool] = field(default=False)
    report_to: List[str] = field(default_factory=lambda: ["tensorboard"])


@dataclass
class DefaultPeftArguments(TrainingArguments):
    output_dir: str
    num_train_epochs: int
    per_device_train_batch_size: int
    gradient_accumulation_steps: int

    load_best_model_at_end: Optional[bool] = field(default=False)
    gradient_checkpointing: Optional[bool] = field(default=True)
    evaluation_strategy: Optional[str] = field(default="no")
    optim: str = field(default="adamw_hf")

    max_grad_norm: Optional[bool] = field(default=0.3)
    weight_decay: float = field(default=0.1)
    lr_scheduler_type: str = field(default="cosine")
    warmup_ratio: float = field(default=0.01)
    logging_steps: int = field(default=10)
    save_strategy: str = field(default="steps")
    save_steps: int = field(default=200)
    save_total_limit: int = field(default=3)
    auto_find_batch_size: Optional[bool] = field(default=False)
    report_to: List[str] = field(default_factory=lambda: ["tensorboard"])

    fp16: Optional[bool] = field(default=False)
    bf16: Optional[bool] = field(default=True)
    neftune_noise_alpha: Optional[int] = field(default=5)


@dataclass
class DefaultLoraConfig(LoraConfig):
    lora_alpha: float = field(default=32)
    lora_dropout: float = field(default=0.1)
    r: int = field(default=64)
    target_modules: List[str] = field(
        default_factory=lambda: [
            "q_proj",
            "v_proj",
            "k_proj",
            "o_proj",
            "gate_proj",
            "up_proj",
            "down_proj",
            "lm_head",
        ]
    )
    task_type: TaskType = field(default=TaskType.CAUSAL_LM)



from typing import Any, Optional

import pandas as pd

from premsql.executors.base import BaseExecutor
from premsql.generators.base import Text2SQLGeneratorBase
from premsql.agents.base import AgentBase, ExitWorkerOutput
from premsql.agents.baseline.workers import (
    BaseLineAnalyserWorker,
    BaseLineFollowupWorker,
    BaseLinePlotWorker,
    BaseLineText2SQLWorker,
)
from premsql.agents.router import SimpleRouterWorker
from premsql.agents.tools.plot.base import BasePlotTool

# TODO: Should the name be changed from baseline to eda or autoeda?


class BaseLineAgent(AgentBase):
    def __init__(
        self,
        session_name: str,
        db_connection_uri: str,
        specialized_model1: Text2SQLGeneratorBase,
        specialized_model2: Text2SQLGeneratorBase,
        executor: BaseExecutor,
        plot_tool: BasePlotTool,
        session_db_path: Optional[str] = None,
        include_tables: Optional[list] = None,
        exclude_tables: Optional[list] = None,
        auto_filter_tables: Optional[bool] = False,
        route_worker_kwargs: Optional[dict] = {},
    ) -> None:
        super().__init__(
            session_name=session_name,
            db_connection_uri=db_connection_uri,
            session_db_path=session_db_path,
            route_worker_kwargs=route_worker_kwargs,
        )
        self.text2sql_worker = BaseLineText2SQLWorker(
            db_connection_uri=db_connection_uri,
            generator=specialized_model1,
            helper_model=specialized_model1,
            executor=executor,
            include_tables=include_tables,
            exclude_tables=exclude_tables,
            auto_filter_tables=auto_filter_tables,
        )
        self.analysis_worker = BaseLineAnalyserWorker(generator=specialized_model2)
        self.plotter_worker = BaseLinePlotWorker(
            generator=specialized_model2, plot_tool=plot_tool
        )
        self.followup_worker = BaseLineFollowupWorker(generator=specialized_model2)
        self.router = SimpleRouterWorker()

    def run(
        self, question: str, input_dataframe: Optional[pd.DataFrame] = None
    ) -> ExitWorkerOutput:
        decision = self.router.run(question=question, input_dataframe=input_dataframe)
        dataframe_from_history = None
        # TODO: This is an assumption that the output tables will be in last
        # 10 conversation

        history_entries = self.history.get(limit=10)
        for entry in history_entries:
            content = entry["message"]
            df = content.show_output_dataframe()
            if df is not None and len(df) > 0:
                dataframe_from_history = content.show_output_dataframe()
                break

        if decision.route_to in ["query", "analyse", "plot"]:
            worker_output = self._execute_worker(
                question=question,
                route_to=decision.route_to,
                input_dataframe=input_dataframe,
                dataframe_from_history=dataframe_from_history,
            )
            exit_output = self._create_exit_worker_output(
                question=question,
                route_taken=decision.route_to,
                worker_output=worker_output,
            )
            if any(
                [
                    exit_output.error_from_analysis_worker,
                    exit_output.error_from_plot_worker,
                    exit_output.error_from_sql_worker,
                ]
            ):
                followup_output = self._handle_followup(exit_output)
                exit_output.followup_suggestion = followup_output.suggestion
                exit_output.followup_route_to_take = (
                    followup_output.alternative_route or "query"
                )  # This is the default route
                exit_output.error_from_followup_worker = (
                    followup_output.error_from_model
                )
        else:
            exit_output = self._handle_followup_route(question=question)
        return exit_output

    def _execute_worker(
        self,
        question: str,
        route_to: str,
        input_dataframe: Optional[pd.DataFrame],
        dataframe_from_history: Optional[pd.DataFrame],
    ):
        decision_mappign = {
            "query": lambda: self.text2sql_worker.run(
                question=question,
                render_results_using="json",
                **self.route_worker_kwargs.get("query", {})
            ),
            "analyse": lambda: self.analysis_worker.run(
                question=question,
                input_dataframe=(
                    dataframe_from_history
                    if input_dataframe is None
                    else input_dataframe
                ),
                **self.route_worker_kwargs.get("analyse", {})
            ),
            "plot": lambda: self.plotter_worker.run(
                question=question,
                input_dataframe=(
                    dataframe_from_history
                    if input_dataframe is None
                    else input_dataframe
                ),
                **self.route_worker_kwargs.get("plot", {})
            ),
        }
        return decision_mappign[route_to]()

    def _create_exit_worker_output(
        self,
        question: str,
        route_taken: str,
        worker_output: Any,  # TODO: change it Literal of worker fixed outputs
    ) -> ExitWorkerOutput:
        exit_output = ExitWorkerOutput(
            session_name=self.session_name,
            question=question,
            route_taken=route_taken,
            db_connection_uri=self.db_connection_uri,
            additional_input=getattr(worker_output, "additional_input", None),
        )
        if route_taken == "query":
            exit_output.sql_string = worker_output.sql_string
            exit_output.sql_reasoning = worker_output.sql_reasoning
            exit_output.sql_output_dataframe = worker_output.output_dataframe
            exit_output.error_from_sql_worker = worker_output.error_from_model

        elif route_taken == "analyse":
            exit_output.analysis = worker_output.analysis
            exit_output.analysis_reasoning = worker_output.analysis_reasoning
            exit_output.analysis_input_dataframe = worker_output.input_dataframe
            exit_output.error_from_analysis_worker = worker_output.error_from_model

        elif route_taken == "plot":
            exit_output.plot_config = worker_output.plot_config
            exit_output.plot_input_dataframe = worker_output.input_dataframe
            exit_output.plot_output_dataframe = worker_output.output_dataframe
            exit_output.image_to_plot = worker_output.image_plot
            exit_output.plot_reasoning = worker_output.plot_reasoning
            exit_output.error_from_plot_worker = worker_output.error_from_model

        return exit_output

    def _handle_followup(self, prev_output: ExitWorkerOutput):
        return self.followup_worker.run(
            prev_output=prev_output,
            db_schema=self.text2sql_worker.db.get_context()["table_info"],
            user_feedback=None,
        )

    def _handle_followup_route(self, question: str) -> ExitWorkerOutput:
        history_entries = self.history.get()
        if len(history_entries) == 0:
            return ExitWorkerOutput(
                session_name=self.session_name,
                question=question,
                route_taken="followup",
                db_connection_uri=self.db_connection_uri,
                additional_input=None,
                followup_suggestion="Before Writing a followup please either query / analyse / plot",
                followup_route_to_take="query",
                error_from_followup_worker=None,
            )
        else:
            followup_output = self.followup_worker.run(
                prev_output=self.history.get(limit=1)[0]["message"],
                user_feedback=question,
                db_schema=self.text2sql_worker.db.get_context()["table_info"],
                **self.route_worker_kwargs.get("followup", {})
            )
        return ExitWorkerOutput(
            session_name=self.session_name,
            question=question,
            route_taken="followup",
            db_connection_uri=self.db_connection_uri,
            additional_input=None,
            followup_suggestion=followup_output.suggestion,
            followup_route_to_take=followup_output.alternative_route
            or "query",  # query should alaways be the default route
            error_from_followup_worker=followup_output.error_from_model,
        )



import subprocess
from typing import Optional

import requests
from api.models import Completions, Session
from api.pydantic_models import (
    CompletionCreationRequest,
    CompletionCreationResponse,
    CompletionListResponse,
    CompletionSummary,
    SessionCreationRequest,
    SessionCreationResponse,
    SessionDeleteResponse,
    SessionListResponse,
    SessionSummary,
)
from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import Paginator
from django.db import transaction

from premsql.logger import setup_console_logger
from premsql.agents.base import AgentOutput
from premsql.agents.memory import AgentInteractionMemory
from premsql.playground import InferenceServerAPIClient
from premsql.playground.backend.api.utils import stop_server_on_port

logger = setup_console_logger("[SESSION-MANAGER]")

# TODO: # When delete a session, then it should delete the memory of the session
# TODO: # when fetching the history it should just give out the message_id in the current django db
# and then using that we can iteratively request the history to give history chats one by one


class SessionManageService:
    def __init__(self) -> None:
        self.client = InferenceServerAPIClient()

    def create_session(
        self, request: SessionCreationRequest
    ) -> SessionCreationResponse:
        response = self.client.get_session_info(base_url=request.base_url)
        if response.get("status") == 500:
            return SessionCreationResponse(
                status_code=500,
                status="error",
                error_message="Can not start session, internal server error. Try Again!",
            )

        try:
            session = Session.objects.create(
                session_name=response["session_name"],
                db_connection_uri=response["db_connection_uri"],
                created_at=response["created_at"],
                base_url=response["base_url"],
                session_db_path=response["session_db_path"],
            )
            logger.info(f"Successfully created session: {response['session_name']}")
            return SessionCreationResponse(
                status_code=200,
                status="success",
                session_id=session.session_id,
                session_name=session.session_name,
                db_connection_uri=response["db_connection_uri"],
                session_db_path=response["session_db_path"],
                created_at=session.created_at,
                error_message=None,
            )
        except Exception as e:
            return SessionCreationResponse(
                status_code=500,
                status="error",
                error_message=f"Can not start session. {e}",
            )

    def get_session(self, session_name: str) -> Optional[Session]:
        try:
            return Session.objects.get(session_name=session_name)
        except ObjectDoesNotExist:
            return None

    def list_session(self, page: int, page_size: int = 20) -> SessionListResponse:
        try:
            sessions = Session.objects.all().order_by("-created_at")
            paginator = Paginator(sessions, page_size)
            page_obj = paginator.get_page(page)
            session_summaries = [
                SessionSummary(
                    session_id=session.session_id,
                    session_name=session.session_name,
                    created_at=session.created_at,
                    base_url=session.base_url,
                    db_connection_uri=session.db_connection_uri,
                    session_db_path=session.session_db_path,
                )
                for session in page_obj
            ]
            return SessionListResponse(
                status="success",
                status_code=200,
                sessions=session_summaries,
                total_count=len(session_summaries),
                page=page,
                page_size=page_size,
            )
        except Exception as e:
            return SessionListResponse(
                status="error",
                status_code=500,
                session_summaries=None,
                total_count=0,
                page=page,
                page_size=page_size,
                error_message=f"Error listing sessions: {e}",
            )

    def delete_session(self, session_name: str):
        try:
            with transaction.atomic():
                session = Session.objects.get(session_name=session_name)
                try:
                    running_port = int(session.base_url.split(":")[1])
                    stop_server_on_port(port=running_port)
                except Exception as e:
                    logger.info(
                        "process killing failed, please shut down inference server manually"
                    )
                    pass

                # Proceed with deletion
                Completions.objects.filter(session_name=session_name).delete()
                session.delete()
                logger.info("Deleted all the chats")
                agent_memory = AgentInteractionMemory(
                    session_name=session_name, db_path=session.session_db_path
                )
                logger.info("Deleted the session registered inside PremSQL Agent")
                agent_memory.delete_table()
                return SessionDeleteResponse(
                    session_name=session_name,
                    status_code=200,
                    status="success",
                    error_message=None,
                )
        except Session.DoesNotExist:
            return SessionDeleteResponse(
                session_name=session_name,
                status_code=404,
                status="error",
                error_message="Session does not exist",
            )
        except Exception as e:
            return SessionDeleteResponse(
                session_name=session_name,
                status_code=500,
                status="error",
                error_message=f"Session does not exist: {e}",
            )


class CompletionService:
    def __init__(self) -> None:
        self.client = InferenceServerAPIClient()

    def completion(
        self, request: CompletionCreationRequest
    ) -> CompletionCreationResponse:
        try:
            session = Session.objects.get(session_name=request.session_name)
        except ObjectDoesNotExist:
            return CompletionCreationResponse(
                status_code=404,
                status="error",
                session_name=request.session_name,
                error_message=f"Session '{request.session_name}' not found",
            )

        try:
            # Small Hack ;_)
            base_url = session.base_url
            base_url = f"http://{base_url}"
            session_inference_response = self.client.post_completion(
                base_url=base_url, question=request.question
            )
        except Exception as e:
            logger.error(f"Unexpected error during completion: {str(e)}")
            return CompletionCreationResponse(
                status_code=500,
                status="error",
                session_name=session.session_name,
                error_message="An unexpected error occurred",
            )

        try:
            chat = Completions.objects.create(
                session=session,
                session_name=session.session_name,
                question=request.question,
                message_id=session_inference_response.get("message_id"),
                created_at=session_inference_response.get("message").get("created_at"),
            )

            logger.info(
                f"Chat completion created successfully for session: {session.session_name}"
            )
            agent_output = AgentOutput(**session_inference_response.get("message"))
            return CompletionCreationResponse(
                status_code=200,
                status="success",
                message_id=chat.message_id,
                session_name=session.session_name,
                created_at=chat.created_at,
                question=chat.question,
                message=agent_output,
            )

        except Exception as e:
            logger.error(f"Error saving completion: {str(e)}")
            return CompletionCreationResponse(
                status_code=500,
                status="error",
                session_name=session.session_name,
                error_message=f"Completion successful, but failed to save: {e}",
            )

    def chat_history(
        self, session_name: str, page: int, page_size: int = 20
    ) -> CompletionListResponse:
        try:
            session = Session.objects.get(session_name=session_name)
        except ObjectDoesNotExist:
            return CompletionListResponse(
                status="error",
                status_code=404,
                completions=[],
                total_count=0,
                page=page,
                page_size=page_size,
                error_message=f"Session '{session_name}' not found",
            )

        try:
            completions = Completions.objects.filter(session=session).order_by(
                "created_at"
            )
            paginator = Paginator(completions, page_size)
            page_obj = paginator.get_page(page)

            completion_summaries = [
                CompletionSummary(
                    message_id=completion.message_id,
                    session_name=completion.session_name,
                    base_url=completion.session.base_url,
                    created_at=completion.created_at,
                    question=completion.question,
                )
                for completion in page_obj
            ]

            return CompletionListResponse(
                status="success",
                status_code=200,
                completions=completion_summaries,
                total_count=completions.count(),
                page=page,
                page_size=page_size,
            )
        except Exception as e:
            return CompletionListResponse(
                status="error",
                status_code=500,
                completions=[],
                total_count=0,
                page=page,
                page_size=page_size,
                error_message=f"Error fetching chat history: {str(e)}",
            )



from typing import Optional

import pandas as pd

from premsql.generators.base import Text2SQLGeneratorBase
from premsql.logger import setup_console_logger
from premsql.agents.base import WorkerBase
from premsql.agents.baseline.prompts import BASELINE_FOLLOWUP_WORKER_PROMPT
from premsql.agents.models import ExitWorkerOutput, FollowupWorkerOutput

logger = setup_console_logger("[BASELINE-FOLLOWUP-WORKER]")


class BaseLineFollowupWorker(WorkerBase):
    def __init__(self, generator: Text2SQLGeneratorBase) -> None:
        self.generator = generator

    def run(
        self,
        prev_output: ExitWorkerOutput,
        db_schema: str,
        user_feedback: Optional[str] = None,
        prompt_template: Optional[str] = BASELINE_FOLLOWUP_WORKER_PROMPT,
        temperature: Optional[float] = 0.18,
        max_new_tokens: Optional[int] = 128,
    ) -> FollowupWorkerOutput:
        if prev_output.route_taken == "query":
            error = "\n".join(
                filter(None, [prev_output.error_from_sql_worker, user_feedback])
            )
            dataframe = prev_output.sql_output_dataframe
        elif prev_output.route_taken == "plot":
            error = "\n".join(
                filter(None, [prev_output.error_from_plot_worker, user_feedback])
            )
            dataframe = prev_output.plot_input_dataframe
        elif prev_output.route_taken == "analyse":
            dataframe = prev_output.analysis_input_dataframe
            error = "\n".join(
                filter(None, [prev_output.error_from_analysis_worker, user_feedback])
            )
        else:
            error = user_feedback
            dataframe = next(
                (
                    df
                    for df in [
                        prev_output.sql_output_dataframe,
                        prev_output.plot_input_dataframe,
                        prev_output.analysis_input_dataframe,
                    ]
                    if df is not None
                ),
                None,
            )

        if dataframe:
            if isinstance(dataframe, dict) and "data" in dataframe and "columns" in dataframe:
                dataframe = pd.DataFrame(dataframe["data"], columns=dataframe["columns"])
            elif not isinstance(dataframe, pd.DataFrame):
                try:
                    dataframe = pd.DataFrame(dataframe)
                except:
                    dataframe = None

        prompt = prompt_template.format(
            schema=db_schema,
            decision=prev_output.route_taken,
            question=prev_output.question,
            dataframe=dataframe,
            analysis=prev_output.analysis,
            error_from_model=error,
        )
        try:
            result = self.generator.generate(
                data_blob={"prompt": prompt},
                temperature=temperature,
                max_new_tokens=max_new_tokens,
                postprocess=False,
            )
            result = eval(result.replace("null", "None"))
            error_from_model = None
            assert "alternate_decision" in result
            assert "suggestion" in result
        except Exception as e:
            result = {
                "alternate_decision": prev_output.route_taken,
                "suggestion": "Worker unable to generate alternative suggestion",
            }
            error_from_model = str(e)

        return FollowupWorkerOutput(
            question=user_feedback or prev_output.question,
            error_from_model=error_from_model,
            route_taken=result["alternate_decision"],
            suggestion=result["suggestion"],
            additional_input={
                "temperature": temperature,
                "max_new_tokens": max_new_tokens,
            },
        )



from datetime import datetime
from typing import Dict, Literal, Optional

import pandas as pd
from pydantic import BaseModel, Field

from premsql.logger import setup_console_logger

logger = setup_console_logger("[BASE-MODELS]")


class BaseWorkerOutput(BaseModel):
    """Base model for worker outputs with common fields."""

    question: str
    error_from_model: Optional[str] = None
    additional_input: Optional[Dict] = Field(
        default=None, description="Additional input data"
    )


class Text2SQLWorkerOutput(BaseWorkerOutput):
    """Output model for Text2SQL worker."""

    db_connection_uri: str
    sql_string: str
    sql_reasoning: Optional[str] = None
    input_dataframe: Optional[Dict] = None
    output_dataframe: Optional[Dict] = None

    def show_output_dataframe(self) -> pd.DataFrame:
        if self.output_dataframe:
            return pd.DataFrame(
                self.output_dataframe["data"], columns=self.output_dataframe["columns"]
            )
        return pd.DataFrame()


class AnalyserWorkerOutput(BaseWorkerOutput):
    """Output model for Analyser worker."""

    analysis: str
    input_dataframe: Optional[Dict] = None
    analysis_reasoning: Optional[str] = None


class ChartPlotWorkerOutput(BaseWorkerOutput):
    """Output model for ChartPlot worker."""

    input_dataframe: Optional[Dict] = None
    plot_config: Optional[Dict] = None
    image_plot: Optional[str] = None
    plot_reasoning: Optional[str] = None
    output_dataframe: Optional[Dict] = None


class RouterWorkerOutput(BaseWorkerOutput):
    """Output model for Router worker."""

    route_to: Literal["followup", "plot", "analyse", "query"]
    input_dataframe: Optional[Dict] = None
    decision_reasoning: Optional[str] = None


# This is a more of a custom worker
class FollowupWorkerOutput(BaseWorkerOutput):
    """Output model for Followup worker."""

    route_taken: Literal["followup", "plot", "analyse", "query"]
    suggestion: str
    alternative_route: Optional[Literal["followup", "plot", "analyse", "query"]] = None


class ExitWorkerOutput(BaseModel):
    """Output model for Exit worker, combining results from all workers."""

    session_name: str
    question: str
    db_connection_uri: str
    route_taken: Literal["plot", "analyse", "query", "followup"]

    # Text2SQL fields
    sql_string: Optional[str] = None
    sql_reasoning: Optional[str] = None
    sql_input_dataframe: Optional[Dict] = None
    sql_output_dataframe: Optional[Dict] = None
    error_from_sql_worker: Optional[str] = None

    # Analysis worker fields
    analysis: Optional[str] = None
    analysis_reasoning: Optional[str] = None
    analysis_input_dataframe: Optional[Dict] = None
    error_from_analysis_worker: Optional[str] = None

    # Plot Worker fields
    plot_config: Optional[Dict] = None
    plot_input_dataframe: Optional[Dict] = None
    plot_output_dataframe: Optional[Dict] = None
    image_to_plot: Optional[str] = None
    plot_reasoning: Optional[str] = None
    error_from_plot_worker: Optional[str] = None

    # Followup Worker fields
    followup_route_to_take: Optional[
        Literal["plot", "analyse", "query", "followup"]
    ] = None
    followup_suggestion: Optional[str] = None
    error_from_followup_worker: Optional[str] = None

    # Additional input
    additional_input: Optional[Dict] = Field(
        default=None, description="Additional input data"
    )

    def show_output_dataframe(
        self,
    ) -> pd.DataFrame:
        dataframe = None
        if self.route_taken == "query":
            dataframe = self.sql_output_dataframe
        elif self.route_taken == "plot":
            dataframe = self.plot_output_dataframe
        elif self.route_taken == "analyse":
            dataframe = self.analysis_input_dataframe

        if dataframe:
            return pd.DataFrame(dataframe["data"], columns=dataframe["columns"])
        return pd.DataFrame()


class AgentOutput(BaseModel):
    """Final output model for the entire pipeline."""

    session_name: str
    question: str
    db_connection_uri: str
    route_taken: Literal["plot", "analyse", "query", "followup"]
    input_dataframe: Optional[Dict] = None
    output_dataframe: Optional[Dict] = None
    sql_string: Optional[str] = None
    analysis: Optional[str] = None
    reasoning: Optional[str] = None
    plot_config: Optional[Dict] = None
    image_to_plot: Optional[str] = None
    followup_route: Optional[Literal["plot", "analyse", "query", "followup"]] = None
    followup_suggestion: Optional[str] = None
    error_from_pipeline: Optional[str] = None
    created_at: datetime = Field(default_factory=datetime.now)

    def show_output_dataframe(
        self,
    ) -> pd.DataFrame:
        dataframe = None
        if self.route_taken == "query":
            dataframe = self.sql_output_dataframe
        elif self.route_taken == "plot":
            dataframe = self.plot_output_dataframe
        elif self.route_taken == "analyse":
            dataframe = self.analysis_input_dataframe

        if dataframe:
            return pd.DataFrame(dataframe["data"], columns=dataframe["columns"])
        return pd.DataFrame()



from premsql.generators.huggingface import Text2SQLGeneratorHF
from premsql.generators.openai import Text2SQLGeneratorOpenAI
from premsql.generators.premai import Text2SQLGeneratorPremAI
from premsql.generators.mlx import Text2SQLGeneratorMLX
from premsql.generators.ollama_model import Text2SQLGeneratorOllama

__all__ = ["Text2SQLGeneratorHF", "Text2SQLGeneratorPremAI", "Text2SQLGeneratorOpenAI", "Text2SQLGeneratorMLX", "Text2SQLGeneratorOllama"]
